#!/bin/sh
##########################################################################
#
# Function: Configure the Makefile and provide a menu for
#           calling of make
#
# Remark:  This is not very usefull here but I need an example
#
# Author:  Jean-Jacques Sarton jj.sarton@t-online.de
#
# Date:    JUN-04-2000
#
# Changes:
#  DEC 21 06  Patch from Paolo regarding the placement of the file
#             selection example.
# DEC  27 o6  Integrated debianization from Paolo with little
#             improvment
#
##########################################################################

### compile xwChoice
[ "$DISPLAY" ] && {
  WxH=`xwininfo -root|grep geometry|sed 's/.* \([0-9]\+x[0-9]\+\).*/\1/'`
  W=${WxH%x*}
  H=${WxH#*x}
  W=${W:-640}
  H=${H:-480}
  W_2=`expr $W / 2`
  H_2=`expr $H / 2`
  W_3=`expr $W / 3`
  H_3=`expr $H / 3`
  W_4=`expr $W / 4`
  H_4=`expr $H / 4`
  W_5=`expr $W / 5`
  H_5=`expr $H / 5`
  echo $WxH $W $H >&2
}

### a few functions for check of installation and
### find out which Xaw libs are installed.

makeTestFile()
{
cat <<!
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>

int main(int argc, char **argv)
{
   XtAppContext appContext;
   Widget toplevel, form;
   toplevel = XtAppInitialize(&appContext,"T",NULL,0,
                              0,NULL,NULL,NULL,0);
   return 0;
}
!
}

makeTestCCFile()
{
cat <<!
#include <stdio.h>
int main() { return 0; }
!
}

checkDevEnv()
{
   case $1 in
   cc)
      makeTestCCFile > tst.c
      $CC -o tst tst.c
      ;;
   *)
      makeTestFile > tst.c
      $CC -o tst tst.c -L/usr/X11R6/lib $1 -lXt -lX11 -I/usr/X11R6/include $2
      ;;
   esac >/dev/null 2>&1
   status=$?
   rm tst tst.c >/dev/null 2>&1
   return $status
}

# find out if we can use gcc or cc
checkForGccOrCC()
{
   for p in `echo $PATH | tr ':' ' '`
   do
      [ -f $p/gcc ] && ( echo gcc; return 0 )
   done
   for p in `echo $PATH | tr '.' ' '`
   do
      [ -f $p/cc ] && ( echo cc; return 0 );
   done
}

if [ "`uname -s`" = Linux ]
then
   if [ -s Makefile ]; then
     echo -e 'Makefile already done - overwrite?[N|y]: \c'
     read r
     case "$r" in
       y|Y|s|j|J|S) rm Makefile ;;
       *) ;;
     esac
   fi

   # general checks
   CC=`checkForGccOrCC`

   if [ -z $CC ]
   then
      echo gcc/cc not found ! Please install development tools
   fi

   if ! checkDevEnv cc
   then
      echo Please install development tools
      exit 1
   fi

   # check for X11/Xt
   if ! checkDevEnv
   then
      echo Please install X11 development files
      exit 1
   fi

   if [ ! -e Makefile ]
   then
      O=
      XAW=
      if checkDevEnv -lXaw; then O=x; XAW="Xaw(x)"; fi
      checkDevEnv -lXaw3d  -DXAW3D=1  && if [ -z $O ]; then O=d; XAW="$XAW Xaw3d(d)";  else O=$O"|d"; XAW=" $XAW Xaw3d(d)";  fi
      checkDevEnv -lneXtaw -DNEXTAW=1 && if [ -z $O ]; then O=n; XAW="$XAW neXtaw(n)"; else O=$O"|n"; XAW=" $XAW neXtaw(n)"; fi
      [ "$XAW" ] && [ "$O" ] || {
        echo 'No *X*aw* widgets found ?!?'
        exit 1
      }

      echo -e "Found these widgets: $XAW - choose: [$O]?:\c "
      read r
      [ -e xwChoice ] && /bin/rm xwChoice
      case "$r" in
        d|D) sed -e 's/^#XAW=3d/XAW=3d/' -e 's/@//' Makefile.org > Makefile ;;
        n|N) sed -e 's/^#XAW=neXt/XAW=neXt/' -e 's/@//' \
           Makefile.org > Makefile ;;
        *) sed -e 's/^#XAW=plain/XAW=plain/' -e 's/@//' \
           Makefile.org > Makefile ;;
      esac
   fi
fi

### set environment
PATH=`pwd`:$PATH
XWCHOICE=`pwd`/xwChoice

#You may set the Fonts to use as follow
#export XwChoiceFn='-cronyx-helvetica-medium-r-normal-*-14-*-*-*-*-*-koi8-r'
#export XwChoiceFb='-cronyx-fixed-medium-r-normal--24-170-100-100-c-120-koi8-r'

### prepare for logging of make output
echo xwChoice Log > make.log
exec 2>&1 1>> make.log
make xwChoice

# debianizing, bail out rightaway...
[ -e debian-build-stamp ] && exit

### I use a few functions so the main code look better and

# xwChoice must be started as background process and calling via
# a function is necessary

doWait()
{
   # start xwChoice as wait box in the background
   $XWCHOICE -title 'Make xwChoice'                         \
             -wait  OK '\n\n\n   Wait a little bit ...\n\n\n' &
}

# Popup the log window if not allready running
showLog()
{
   if [ "$ProcessId" = "" -o ! -d /proc/$ProcessId ]
   then
      $XWCHOICE -title 'Make xwChoice'     \
                -geometry +$W_5+$H_5             \
                -tail 'close Log viewer' make.log &
      ProcessId=$!
   fi
}

# only in order to have a message file which can be outputed
bldHlp()
{
   if [ -f helpFile ]
   then
      return
   fi
   echo 'This is an example for an help text
This is the second line
and the third.

Enjoy with xwChoice' > helpFile
}

# a simple file selection box
fsb()
{
   wd=`pwd`
   title='File Selection Box'
   question="actual dir:\n$wd"
   back='Done'
   ok='Select'

   # build list of directories and files
   dirs="../ `ls -F1 | grep / | sort`"
   files="`ls -F1 | grep -v / | tr -d '*=|@' | sort`"
   # make a script with all items
   {
      echo "xwChoice -title '$title'       \\"
      echo "         -question '$question' \\"
      echo "         -cancel   '$back' ''  \\"
      echo "         -geometry +$W_3+$H_3  \\"
      echo "         -columns 3 -lines 24  \\"

     for d in $dirs
     do
        echo "       -choice '$d' '$d' \\"
     done
     for f in $files
     do
        echo "       -choice '$f' '$f' \\"
     done
     echo "          -ok '$ok'"
   } > /tmp/fsb.$$

   # make the script executable   
   chmod +x /tmp/fsb.$$

   answer=`/tmp/fsb.$$`
   choice=$wd/$answer
   status=$?
   rm /tmp/fsb.$$

   # Cancel choosen -> we have our choice now
   if [ $status -eq 3 ]
   then
      rm /tmp/fsb.$$
      $XWCHOICE -error ok "You choosed\n$choice\n"
      return;
   fi

   # special case for links
   if [ -h $wd/$answer -a -d $wd/$answer ]
   then
      answer=$answer/
   fi

   # perform action according to entry type
   case $answer in
      '../') cd $wd/..; wd=`pwd`;fsb;;
      */)    cd $wd/$answer; fsb;;
      *)     $XWCHOICE -message OK "`file $wd/$answer`";; 
   esac
}

# Demonstation for an input field
input()
{
   answer=`$XWCHOICE -question 'Enter some text:' -Input -ok Done`
   $XWCHOICE -message OK "You have entered:\n\n$answer\n\n"
}

showFont()
{
   # make a little message file
   echo 2 3 4 5 6 7 10 11 12 13 14 15 | awk 'BEGIN{ RS=" "} { 
      for ( i = 0; i < 16; i++ ) printf("%c",$0 * 16 + i)
      printf("\n")
   }' > AllChars.txt

   for  s in 8 10 16 24
   do  
      $XWCHOICE -title "--- FONTS $s points ---" \
                -message  OK AllChars.txt \
                -fn      "-*-helvetica-medium-r-normal-*-$s-*-*-*-*-*-*-*" \
                -ok        'OK'
   done
   rm AllChars.txt
}

multiChoice()
{
   selected=`$XWCHOICE -question  'Choose an item:'    \
                       -columns    3                   \
                       -choice    'item A'    'item A' \
                       -choice    'item B'    'item B' \
                       -choice    'item C'    'item C' \
                       -choice    'item D'    'item D' \
                       -choice    'item E'    'item E' \
                       -choice    'item F'    'item F' \
                       -choice    'item G'    'item G' \
                       -defChoice 'item E'             \
                       -Input                          \
                       -cancel    'Cancel'    quit     \
                       -ok        'OK'`

   $XWCHOICE -message OK "your selection is:\n$selected" 
}

# login and password
getLogin()
{
   while :
   do
      loginPasswd=`$XWCHOICE -title  'Enter user name and password' \
                             -login  'User Name:'                   \
                             -passwd 'Password:'                    \
                             -cancel 'Cancel' cancel                \
                             -ok     'OK'`

      Login=`echo $loginPasswd | sed 's/\(.*\):.*/\1/'`
      Passwd=`echo $loginPasswd | sed 's/.*:\(.*\)/\1/'`
      if [ "$loginPasswd" = cancel ]
      then
         $XWCHOICE -title 'Login' \
                   -message OK 'Login cancelled !'
         break
      fi

      if [ "$Login" != "root" ]
      then
         $XWCHOICE -title 'Login' \
                   -error OK 'Please enter root as login name'
         continue
      fi

      if [ "$Passwd" != "root" ]
      then
         $XWCHOICE -title 'Passwd' \
                   -error OK 'Please enter root as password'
         continue
      fi
      $XWCHOICE -title 'Login' \
                -message OK 'Congratulation !'
      break;
   done
}

merchandises()
{
   $XWCHOICE -title     'xwChoice'                      \
             -question  'select all wanted merchandise' \
             -check     'Bananas' Bananas               \
             -check     'Apples'  Apples                \
             -check     'Oranges' Oranges
}

getMerchandises()
{
   merchandises=`merchandises`
   case $merchandises in
   cancel) $XWCHOICE -message OK 'No merchandises choosed';;
   *)
      {
         echo Your choice:
         echo
         echo $merchandises| tr ':' '\012'
      } | $XWCHOICE -tail OK -
      ;;
   esac 
}

oneoff()
{
   result=`$XWCHOICE -title     'xwChoice'           \
                     -question  'select wanted item' \
                     -check     'item a' a           \
                     -check     'item b' b           \
                     -check     'item c' c           \
                     -radio`
    case $result in
    cancel) $XWCHOICE -title 'xwChoice' -error OK 'Canceled';;
    "") $XWCHOICE -title 'xwChoice' -error OK 'No Choice !';;
    *) $XWCHOICE -title 'xwChoice' -message OK "Your choice: Item $result";;
    esac
}

# our main menu
menu()
{
   $XWCHOICE -title     'Make xwChoice'                                   \
             -lines     13                                                \
             -question  'Choose the action you want:'                     \
             -choice    'show me the documentation'             doc       \
             -choice    'Remove all temporary files'            clean     \
             -choice    'Build xwChoice'                        xwChoice  \
             -choice    'Install xwChoice to /usr/local/bin'    install   \
             -choice    'show error message'                    errMsg    \
             -choice    'File Selection'                        fsb       \
             -choice    'Input Box'                             input     \
             -choice    'Login Box'                             login     \
             -choice    'show Font'                             showFont  \
             -choice    'multi columns choices'                 choices   \
             -choice    'multi choices'                      merchandises \
             -choice    'One Off'                               oneoff    \
             -defChoice 'Build xwChoice'                                  \
             -cancel    'Quit'                                  quit      \
             -help      'Help me' 'I have read the Help text' helpFile    \
             -ok        'Do the hard job'
}
#             -help      'Help me' 'I have read the Help text' helpFile    \

### begin with the job
ProcessId=

while :
do
   bldHlp  # build help file needed by menu
   answer=`menu`
   case $answer in
      doc)      $XWCHOICE -title 'Make xwChoice' -lines 25 -log OK xwChoice.txt;;
      clean)    doWait; sleep 1; rm *.o;;
      xwChoice) showLog; make $XWCHOICE ;;
      install)  cp $XWCHOICE /usr/local/bin;;
      errMsg)   $XWCHOICE -error OK '  An error\noccured here';;
      fsb)      curDir=`pwd`
                cd /
                fsb
                cd $curDir;;
      input)    input;;
      login)    getLogin;;
      choices)  multiChoice;;
      showFont) showFont;;
      merchandises)getMerchandises;;
      oneoff)   oneoff;;
      quit)     rm helpFile make.log; exit;;
   esac
done

