/*
*********************************************************************
**
** Copyright (C) jjsa 1999-2000  All Rights Reserved  Confidential
**
*********************************************************************
*/

static char xwChoice_Vxc[] =
                    "@(#)xwChoice.c V1.14 2006-DEC-21/jjsa (C) jjsa - GPL";
/*
 *********************************************************************
 *
 *  MODULE:      xwChoice
 *
 *  CONTENTS:
 *
 *  FUNCTION:   A simple choice box for installing of the
 *              xw_tools
 *
 *  CREATION:   jjsa 08.AUG.1999
 *
 *  CHANGES:    jjsa 04.JAN.2000 deleted <Return> and <Tab> from text
 *                               widget input. <Space> conditional
 *                               opt -input no Space allowed,
 *                                   -Input Space allowed
 *                               opt -deny for text input added
 *              jjsa 20.FEB.2000 opt -defChoice added
 *              jjsa 29.APR-2000 added handling from keyboard
 *              jjsa 07.MAI-2000 focus change on mouse event better handled
 *              jjsa 19-MAI-2000 Added file viewer (-log ok file)
 *              jjsa 27-MAI-2000 Added file viewer (-tail ok file)
 *                               review of handling of scrolled list
 *              jjsa 10-JUN-2000 replaced AddInput by AddTimeOut
 *                               XtAppAddInput() work not as expected
 *                               CPU usage is very high
 *              jjsa 17-JUN-2000 psuhbitton not correctly labeled
 *                               for the log / tail window, corrected
 *              jjsa 23-JUN-2000 List accept <return> as OK
 *              paolo 03/04/06   fixed bug: lookup X-resources file according
 *                               to fixed program name in 'program' not the
 *                               contents of 'title'.
 *                               Added brief usage() and corresponding
 *                               -h|--help
 *                               Say something on usage error.
 *              jjsa 09-APR-2006 remanied a little bit the above bug fix
 *              paolo 10/04/06   - <return|enter> in inText call okCB as well
 *                               - dont exit(2) if no -question but have
 *                                 either -input | -choice
 *                               - usage() takes an arg: 1 = print err as well
 *                               - defaults for -question
 *                               - dont exit(2) on missing args to -opts if
 *                                 a default is defined.
 *                               - logwnd: won't show more than lines*cols,
 *                                 hence dont malloc&suck in whole logFile,
 *                                 just enough.
 *              jjsa 24-APR-2006 - reworked layout function with check of
 *                                 size and use of text widget instead of
 *                                 labels (use of scrollbars if the needed
 *                                 window size is to big).
 *                                 Most resources may now be set within
 *                                 a resource file.
 *              paolo   -APR-2006  some little bugs corected and review
 *                                 of some resources.
 *              jjsa  01-MAI-2006  reviewed the code according to request
 *                                 from paolo.
 *                                 adder new private resources.
 *              paolo   -MAI-2006  a few adds and bug fixed
 *              jjsa  03-Mai-2006  Scrolling within the list implemented
 *                                 double click within the list act as "OK"
 *                                 button handlers reviewed.
 *              jjsa  25-Mai-2006  A lot of little changes concerning layout,
 *                                 ...
 *              jjsa  11-JUL-2006  Added login box
 *              jjsa  24-JUL-2006  Added check box
 *              jjsa  24-AUG-2006  Added location resource in order to place
 *                                 the windows on the top, bottom, right
 *                                 and or left.
 *              jjsa  11-OCT-2006  Xaw3d shown not the last line (tail option)
 *              jjsa  07-OCT-2006  Fluxbox and also other WM don't set the
 *                                 width and height of the shell, code
 *                                 modified.
 *                                 refresh problem with text widget fixed
 *              jjsa  12-DEC-2006  corrected argument parsing and printout for
 *                                 wrong arguments
 *              jjsa  13-DEC-2006  corrected wrong dimensioned array as
 *                                 reported by paolo.
 *                                 cursor not visible within login
 *                                 window (reported by paolo) fixed.
 *                                 updated usage() with new functions
 *
 *              jjsa  21-DEC-2006  login/password window remanied.
 *                                 (caret handling).
 *              jjsa  22-DEC-2006  Handling aste into password widget added.
 *
 *********************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/stat.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xatom.h>

#include <X11/keysym.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Text.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Porthole.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Toggle.h>

#define GREY "grey90"
#define GREY85 "grey85"
#define GREY95 "grey96"
#ifdef XAW3D
#define WITHXAW "Compiled with Xaw3D"
#else
# ifdef NEXTAW
# define WITHXAW "Compiled with neXtaw"
# else
# define WITHXAW "Compiled with Xaw"
# endif
#endif

#define XWCHOICE_KB    1
#define XWCHOICE_MOUSE 2
#define HSPACE 10
#define VSPACE 10

#define SCRBW 14

char errorText[1024];

/* our own resources */
typedef struct appResourceRec_s
{
   Dimension  minButtonWidth;      /* if 0 don't apply, size as required by label */
   char      *buttonAlignment;     /* Left, Right or Center */
   int        topMargin;
   int        rigthMargin;
   int        leftMargin;
   int        bottomMargin;
   int        center;              /* center window */
   int        doubleKlickDelay;
   char      *location;
} appResourceRec_t;

appResourceRec_t appResourceRec;

XtResource resources[] = {

  {"minButtonWidth", "minButtonWidth", XtRDimension, sizeof(Dimension),
    XtOffsetOf(appResourceRec_t, minButtonWidth), XtRImmediate,  (XtPointer) 50 },

  {"topMargin", "topMargin", XtRPosition, sizeof(Position),
    XtOffsetOf(appResourceRec_t, topMargin), XtRImmediate,       (XtPointer) 65 },

  {"rigthMargin", "rigthMargin", XtRPosition, sizeof(Position),
    XtOffsetOf(appResourceRec_t, rigthMargin), XtRImmediate,     (XtPointer)  0 },

  {"leftMargin", "leftMargin", XtRPosition, sizeof(Position),
    XtOffsetOf(appResourceRec_t, leftMargin), XtRImmediate,   (XtPointer)  0 },

  {"bottomMargin", "bottomMargin", XtRPosition, sizeof(Position),
    XtOffsetOf(appResourceRec_t, bottomMargin), XtRImmediate, (XtPointer) 50 },

  {"center", "center", XtRBool, sizeof(int),
    XtOffsetOf(appResourceRec_t, center), XtRImmediate, (XtPointer) 1 },

  {"buttonAlignment", "buttonAlignment", XtRString, sizeof(String),
    XtOffsetOf(appResourceRec_t, buttonAlignment), XtRImmediate, (XtPointer) "Center" },

  {"doubleKlickDelay", "doubleKlickDelay", XtRInt, sizeof(int),
    XtOffsetOf(appResourceRec_t, doubleKlickDelay), XtRImmediate, (XtPointer) 250 },

  {"location", "location", XtRString, sizeof(String),
   XtOffsetOf(appResourceRec_t, location), XtRImmediate, (XtPointer) "center" },
};

String fallback[] =
{
   "XwChoice.center: 1",
   "XwChoice.location: center",
#if 0
    /* these resource are set above. If you
     * want to set them externally put
     * following into your resource file
     */
   "XwChoice,minButtonWidth:   50",
   "XwChoice,topMargin:        65",
   "XwChoice,rigthMargin:       0",
   "XwChoice,leftMargin:     0",
   "XwChoice,bottomMargin:  50",
   "XwChoice,center:         true",
   "XwChoice,bottomMargin: 300",
#endif
   "XwChoice*background: "GREY,
   "XwChoice*label.background: "GREY95,
   "XwChoice*font:  *-courier-medium-r-normal--10-*",
   "XwChoice*label.font: *-helvetica-bold-r-normal--14-*",
   "XwChoice*Command.borderWidth: 1",
   "XwChoice*Command.font: *-helvetica-bold-r-normal--10-*",
   "XwChoice*Command.background: "GREY85,
   "XwChoice*text.borderWidth: 1",
   "XwChoice*text.borderColor: black",
   "XwChoice*text*background: "GREY95,
   "XwChoice*text.foreground: black",


   /* log window */
   "XwChoice.logForm.label.foreground: blue",
   "XwChoice*logForm.vp.background: white",

   /* login window */
   "XwChoice*loginForm.text.background: white",

   /* choice window */
   "XwChoice.choiceForm.label.foreground: darkgreen",
   "XwChoice.choiceForm*list.background: white",
   "XwChoice.choiceForm*list.borderWidth: 1",
   "XwChoice.choiceForm*vp.borderWidth: 1",
   "XwChoice*vp*shadowWidth: 0",

   /* error window */
   "XwChoice*error.text*hScrollbar.background: "GREY,
   "XwChoice*error.text*vScrollbar.background: "GREY,

   "XwChoice*error.text*background:  yellow1",
   /* need text*font and not text.font ! */
   "XwChoice*error.text*font: *-courier-bold-r-normal--10-*",

   /* Toggle button */
   "XwChoice*Toggle.background: white",
   "XwChoice*Toggle.foreground: green",
   /* help window */
#if 0
   /* special resource for Xaw3d */
   "XwChoice*error*vScrollbar*shadowWidth:  2",
   "XwChoice*error*hScrollbar*shadowWidth:  2",
   "XwChoice*error.okButton.shadowWidth:  2",
   "XwChoice*error*shadowWidth:  0",
   "XwChoice*message*vScrollbar*shadowWidth:  2",
   "XwChoice*message*hScrollbar*shadowWidth:  2",
   "XwChoice*message.okButton.shadowWidth:  2",
   "XwChoice*message*shadowWidth:  0",
   "XwChoice*helpForm*vScrollbar*shadowWidth:  2",
   "XwChoice*helpForm*hScrollbar*shadowWidth:  2",
   "XwChoice*helpForm.okButton.shadowWidth: 2",
   "XwChoice*helpForm*shadowWidth:  0",
#endif
   NULL
};

typedef struct choice_s
{
   char **text;
   char **ret;
   int    count;
} choice_t;

typedef struct check_s
{
   Widget box;
   Widget label;
   char  *text;
   char  *ret;
   Dimension width;
   struct check_s *next;
} check_t;

choice_t  choice    = { NULL, NULL, 0 };
check_t  *check     = NULL;
char     *defCheck  = NULL;
char     *title     = NULL;
char     *help      = NULL;
char     *helpFile  = NULL;
char     *cancel    = "Cancel";
char     *cancelRet = "cancel";
char     *ok        = "OK";
char     *helpOK    = "OK";
char     *logText   = NULL;
char     *loginTxt  = NULL;
char     *passwdTxt = NULL;
char     *question  = NULL;
char     *error     = NULL;
char     *errFile   = NULL;
char     *message   = NULL;
char     *tail      = NULL;
char     *Wait      = NULL;
char     *columnsT  = NULL;
char     *linesT    = "10";
int       lines;

int       input     = 0;
int       Input     = 0;
int       both      = 0;
int       noReturn  = 0;
char     *defChoice = NULL;
char     *class     = "XwChoice";
char     *program   = "xwChoice";
int       radio     = 0;

Dimension screenH   = 0;
Dimension screenW   = 0;

Widget    mf;
Widget    tlb;
Widget    list;
Widget    quest;
Widget    cancelB;
Widget    helpB;
Widget    okB;
Widget    helpW;
Widget    helpM;
Widget    msgT;
Widget    inText;
Widget    vp = NULL;
Widget    loginIn  = NULL;
Widget    passwdIn = NULL;
char     *passwdInput = NULL;
//char     *passwdFake  = NULL;

Widget    selectedWidget;

char      geoBuf[100];
Pixel     white, darkblue, grey, red, dgrey;
Dimension viewPortHeight = 0xffff;
Dimension viewPortWidth  = 0xffff;
XtAppContext appContext;

Widget   toplevel;
Widget   hlplevel;

char    *display  = ":0.0";
char    *geometry =  NULL;
char    *fn       =  NULL;
char    *fb       =  NULL;

int logLines   = 0;
int logColumns = 0;
int cols       = 80;

#if 0
String messageTranslation = {
   "<Key>Down: scroll-one-line-up() \n"
   "<Key>Up:   scroll-one-line-down() \n"
   "<Key>Page_Up:    previous-page() \n"
   "<Key>KP_Page_Up: previous-page() \n"
   "<Key>Page_Down:    next-page() \n"
   "<Key>KP_Page_Down: next-page() \n"
   "<Key>Home:    beginning-of-file() \n"
   "<Key>KP_Home: beginning-of-file() \n"
   "<Key>End:    end-of-file() \n"
   "<Key>KP_End: end-of-file() \n"
   "<Btn5Down>:  scroll-one-line-up() \n"
   "<Btn4Down>:  scroll-one-line-down() \n"
};
#endif

String textTranslation = {
   "<Key>Return: \n<Key>Tab: \n<Key>\\ : "
};

String textTranslationSpc = {
   "<Key>Return: \n<Key>Tab: "
};

char *denyChars = NULL;

String buttonTranslation = {
   "<Key>Return:  set() notify() unset()\n"
   "<Key>KP_Enter:  set() notify() unset()\n"
   "<KeyUp>Tab: highlight()\n<KeyDown>Tab: unhighlight()\n"
   "<Key>0x20:  set() notify() unset()"
};

String toggleTranslation = {
   ""
};

enum {
   LABEL,
   QUESTION,
   VIEWPORT,
   LIST,
   TEXT,
   CANCEL,
   HELP,
   OKBUTTON
} dim_e;

typedef struct dim_s
{
   Dimension w;   /* width */
   Dimension bw;  /* border width */
   Dimension h;   /* height */
   Dimension tw;  /* total width, calculated */
} dim_t;

typedef struct widTab_s
{
   struct widTab_s *next;
   struct widTab_s *prev;
   Widget           wid;
} widTab_t;
widTab_t *widgetTab = NULL;

int    ovalButton  = 0;

void setFocus_CB(Widget w, XtPointer closure, XEvent *event,
                 Boolean *contDisp);
void ListSel_CB(Widget w, XtPointer closure, XEvent *event,
                Boolean *contDisp);
void okCB(Widget widget, XtPointer clientData, XtPointer callData);
int calcTextSize(Widget wid, char *text, int *lines, int *columns,
                  Dimension *height, Dimension *width);
void setWidth(Widget w, Dimension width);
void scrollMsg_CB(Widget w, XtPointer closure, XEvent *event, Boolean *contDisp);
void setWidthAndHeight(Widget w, Dimension width, Dimension height);
void getText(char *name, char **text);
static void getSpacing(Widget w, Dimension *hor, Dimension *vert);
static void calcHeight(Widget w, Dimension vSpace, Dimension *newHeight);
void setWidthAndX(Widget w, Dimension width, Dimension x);
void passwdCB(Widget widget, XtPointer clientData, XtPointer callData);
Widget createLabel(Widget parent, char *name, char *text, Widget top);

#define IF_OPT(a,b)  if (  strcmp(*argv, a) ==  0 )\
                      {\
                         argc--;\
                         argv++;\
                         if ( argc > 0 )\
                         {\
                            b = *argv;\
                            argc--;\
                            argv++;\
                         }\
                         else\
                         {\
                            snprintf(errorText, sizeof(errorText)," option %s\n", a);\
                            usage(1, errorText);\
                            exit(2);\
                         }\
                      }

#define IF_OPTC(a,b)  if (  strcmp(*argv, a) ==  0 ) \
                       {\
                         argc--;\
                         argv++;\
                         if ( argc > 1 )\
                         {\
                            addChoice(argv);\
                            argc -= 2;\
                            argv += 2;\
                         }\
                         else\
                         {\
                            snprintf(errorText, sizeof(errorText)," option %s\n", a);\
                            usage(1, errorText);\
                            exit(2);\
                         }\
                       }

#define IF_OPTCc(a,b)  if (  strcmp(*argv, a) ==  0 ) \
                       {\
                         argc--;\
                         argv++;\
                         if ( argc > 1 )\
                         {\
                            addCheck(argv);\
                            argc -= 2;\
                            argv += 2;\
                         }\
                         else\
                         {\
                            snprintf(errorText, sizeof(errorText)," option %s\n", a);\
                            usage(1, errorText);\
                            exit(2);\
                         }\
                       }

#define IF_OPTB(a,b,c)  if (  strcmp(*argv, a) ==  0 ) \
                        {\
                         argc--;\
                         argv++;\
                         if ( argc > 0 )\
                         {\
                            b = *argv; \
                            argv++; \
                            argc--;\
                            if ( argc > 0 )\
                            {\
                              c = *argv; \
                              argv++;\
                              argc --;\
                            }\
                            else\
                            {\
                              snprintf(errorText, sizeof(errorText)," option %s\n", a);\
                              usage(1, errorText);\
                              exit(2);\
                            }\
                         }\
                         else\
                         {\
                            snprintf(errorText, sizeof(errorText)," option %s\n", a);\
                            usage(1, errorText);\
                            exit(2);\
                         }\
                       }


#define IF_OPTH(a,b,c,d)  if (  strcmp(*argv, a) ==  0 ) \
                        {\
                         argc--;\
                         argv++;\
                         if ( argc > 2 )\
                         {\
                            b = *argv; \
                            argv++; \
                            c = *argv; \
                            argv++;\
                            d = *argv; \
                            argv++;\
                           argc -= 3;\
                         }\
                         else\
                         {\
                            snprintf(errorText, sizeof(errorText)," option %s\n", a);\
                            usage(1, errorText); \
                            exit(2);\
                         }\
                       }

#define IF_OPT1(a,b)  if (  strcmp(*argv, a) ==  0 ) \
                        {\
                         argc--;\
                         argv++;\
                         b = 1; \
                       }

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void addWidgetTab
 *
 *  FUNCTION:   add a widget to the widget tab list, this will allow
 *              us to navigate to the wanted widget via the tab key
 *
 *  INPUT:      Widget w
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/
void addWidgetTab(Widget w)
{
   widTab_t *n;
   widTab_t *p;
   if ( widgetTab == NULL )
   {
      widgetTab = (widTab_t*)calloc(sizeof(widTab_t),1);
      n = widgetTab;
      p = NULL;
   }
   else
   {
      n = widgetTab;
      while ( n->next != NULL )
      {
         n = n->next;
      }
      n->next = (widTab_t*)calloc(sizeof(widTab_t),1);
      p = n;
      n = n->next;
   }
   if ( n )
   {
      n->prev = p;
      n->wid = w;
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void prtText
 *
 *  FUNCTION:   append text into log widget if there is an input
 *
 *  INPUT:      Widget w
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/
static int     size = 0;

#define USE_TIMER 1
#ifdef USE_TIMER
#define BUF_SIZE 8092
void prtText(XtPointer closure, XtInputId *id)
#else
#define BUF_SIZE 1024

void prtText(XtPointer closure, int *source, XtInputId *id)
#endif
{
   char           buffer[BUF_SIZE+1];
   XawTextBlock   tB;
   static char   *text = NULL;
   char          *s = NULL;
   char          *t = NULL;
   static int     got  = 0;
   int            line = 0;
   XawTextPosition top = 0;
   char          *txt;

   buffer[BUF_SIZE] = '\0';
   if ( text == NULL )
   {
      text = (char *)calloc(1,1);
   }
#ifdef USE_TIMER
   if ( (got = read((int)closure, buffer, sizeof(buffer)-1)) > 0)
#else
   if ( (got = read(*source, buffer, sizeof(buffer)-1)) > 0)
#endif
   {
      buffer[got] = 0;
      /* calculate the number of new lines */
      s = t = buffer;

      while ( s && *s )
      {
         s = strchr(s, '\n');
         if ( s )
         {
            logLines++;
            logColumns = logColumns > (s-t) ? logColumns : (s-t);
            s++;
            t = s;
         }
      }

      tB.firstPos = 0;
      tB.length   = got;
      tB.ptr      = buffer;
      tB.format   = 0;

      /* Xaw3d don't show the last line */
      if ( size == 0 )
      {
         tB.length++;
         buffer[got]='\n';
         buffer[got+1]=0;
      }

      XtVaSetValues(inText, XtNeditType, XawtextEdit, NULL);
      XawTextSetInsertionPoint(inText,size);
      XawTextReplace(inText, size, size+got,&tB);

      XawTextSetInsertionPoint(inText,size);
      XtVaGetValues(inText, XtNstring, &txt, NULL);
      size += got;

      if (logLines > lines && txt)
      {
          line = lines;
          s = &txt[size-1];
          {
             while (s > txt && line <= lines )
             {
                 while ( s > txt && *s != '\n' )
                 {
                    s--;
                 }
                 if ( *s == '\n' )
                 {
                    s--;
                 }
                 line++;
             }
             top = s-txt;
         }
      }

      XawTextSetInsertionPoint(inText,top);
      XtVaSetValues(inText, XtNeditType, XawtextRead, NULL);

      if ( logColumns > cols )
          XtVaSetValues(inText,
                        XtNscrollHorizontal, XawtextScrollAlways,
                        NULL);

      if ( logLines > lines )
      {
          XtVaSetValues(inText,
                        XtNscrollVertical, XawtextScrollAlways,
                        NULL);
          XtRemoveEventHandler(inText,
                               KeyPressMask|KeyReleaseMask|ButtonReleaseMask|ButtonPressMask|PointerMotionMask|ButtonMotionMask,
                               True,
                               (XtEventHandler)scrollMsg_CB, (XtPointer)False);
          XtAddEventHandler(inText,
                            KeyPressMask|KeyReleaseMask|ButtonReleaseMask|ButtonPressMask|PointerMotionMask|ButtonMotionMask,
                            True,
                            (XtEventHandler) scrollMsg_CB, (XtPointer)True);
      }
      /* refresh (problem with Xaw3D */
      XawTextInvalidate(inText,0, size);
   }
#ifdef USE_TIMER
   XtAppAddTimeOut(appContext,10, prtText, closure);
#endif
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void passwdCB
 *
 *  FUNCTION:   return login and password according to
 *              configuration
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void checkOkCB(Widget widget, XtPointer clientData, XtPointer callData)
{
   check_t *c = check;
   int set = 0;
   Boolean state;

   while (c )
   {
      XtVaGetValues(c->box, XtNstate, &state, NULL);
      if ( state )
      {
         if ( set )
            printf(":");
         printf("%s",c->ret?c->ret:c->text);
         set = 1;
      }
      c = c->next;
   }
   printf("\n");
   exit(0);
}


/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void getText
 *
 *  FUNCTION:   read text from file or string
 *
 *  INPUT:      char   *name
 *
 * OUTPUT:      char  **text
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void getText(char *name, char **text)
{
   struct stat  stat;
   FILE        *fp;
   char        *s, *t;

   if ( name && (fp = fopen(name,"r")) )
   {
      /* create a new window an print the message */
      /* we use a dialog window                   */
      fstat(fileno(fp),&stat);
      *text = (char*)malloc(stat.st_size+1);
      if ( *text )
      {
         fread(*text, stat.st_size, 1,fp);
         (*text)[stat.st_size] = '\0';
      }
      else
      {
        fprintf(stderr,"%s: can't alloc memory for file=%s\n",
                program, name);
        exit(1);
      }
      fclose(fp);
   }
   else
   {
      *text = name;
      t = s = *text;
      while(*s)
      {
         if ( *s == '\\' && s[1] == 'n' )
         {
            *t++ = '\n';
            s += 2;
         }
         else
            *t++ = *s++;
      }
      *t = '\0';
   }

}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       createContainer
 *
 *  FUNCTION:   create the top container
 *
 *  INPUT:      Widget  shell
 *              char   *name
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/


Widget createContainer(Widget shell, char *name)
{
   int n;
   Arg args[5];

   n = 0;
   return XtCreateManagedWidget(name,
                                formWidgetClass,
                                shell,args,n);
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       toggleCB
 *
 *  FUNCTION:   toggle the checkbox widgets (radio behaviour)
 *              this insure that at last one of the check box
 *              is selected.
 *
 *  INPUT:      Widget  w
 *              XtPointer  data1, data2  unused
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/
void toggleCB(Widget w, XtPointer data1, XtPointer data2)
{
   check_t *c = check;
   while ( c )
   {
      if ( c->box == w )
      {
         XtVaSetValues(c->box, XtNstate, 1, NULL);
      }
      else
      {
         XtVaSetValues(c->box, XtNstate, 0, NULL);
      }
      c = c->next;
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       Widget createCheck
 *
 *  FUNCTION:   create check entry
 *
 *  INPUT:      Widget  parent
 *              Widget  top
 *              char   *text
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

Widget createCheck(Widget parent, Widget top, check_t *check)
{
   int n;
   Arg args[20];
   Position  x;
   Dimension w;

   n = 0;
   XtSetArg(args[n], XtNfromVert,           top);n++;
   XtSetArg(args[n], XtNlabel,              " ");n++;
   XtSetArg(args[n], XtNtop,                XtChainTop);n++;
   XtSetArg(args[n], XtNbottom,             XtChainTop);n++;
   XtSetArg(args[n], XtNleft,               XtChainLeft);n++;
   XtSetArg(args[n], XtNright,              XtChainLeft);n++;
   XtSetArg(args[n], XtNborderWidth,        2);n++;
   XtSetArg(args[n], XtNhighlightThickness, 0);n++;
   XtSetArg(args[n], XtNshapeStyle,         XawShapeRoundedRectangle);n++;
   XtSetArg(args[n], XtNcornerRoundPercent, 0);n++;

   check->box = XtCreateManagedWidget("check",
                                      toggleWidgetClass,
                                      parent,
                                      args,n);

   n = 0;
   XtSetArg(args[n], XtNtop,           XtChainTop);n++;
   XtSetArg(args[n], XtNbottom,        XtChainTop);n++;
   XtSetArg(args[n], XtNfromVert,      top);n++;
   XtSetArg(args[n], XtNleft,          XtChainLeft);n++;
   XtSetArg(args[n], XtNfromHoriz,     check->box);n++;
   XtSetArg(args[n], XtNright,         XtChainRight);n++;
   XtSetArg(args[n], XtNborderWidth,   0);n++;
   XtSetArg(args[n], XtNinternalWidth, 0);n++;
   XtSetArg(args[n], XtNlabel,         check->text); n++;
   XtSetArg(args[n], XtNjustify,        XtJustifyLeft); n++;
   check->label = XtCreateManagedWidget("clabel",
                                        labelWidgetClass,
                                        parent, args, n);
   XtVaGetValues(check->label,
                 XtNx,     &x,
                 XtNwidth, &w,
                 NULL);
   if ( radio )
   {
      XtAddCallback(check->box, XtNcallback, toggleCB, (XtPointer)check);
   }
   check->width = x + w;
   return check->box;
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       Widget createLabel
 *
 *  FUNCTION:   create a text widget
 *
 *  INPUT:      Widget       parent
 *              char        *name
 *              char        *text
 *              int          editable
 *              int          caret     show caret
 *              XFontStruct *font
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/
Widget createLabel(Widget parent, char *name, char *text, Widget top)
{
   Widget w;
   Arg    args[20];
   int    n = 0;
   
   XtSetArg(args[n], XtNtop,      XtChainTop);n++;
   XtSetArg(args[n], XtNbottom,   XtChainTop);n++;
   XtSetArg(args[n], XtNfromVert, top);n++;
   XtSetArg(args[n], XtNleft,     XtChainLeft);n++;
   XtSetArg(args[n], XtNright,    XtChainLeft);n++;
   XtSetArg(args[n], XtNborderWidth,   0);n++;
   XtSetArg(args[n], XtNinternalWidth, 0);n++;
   XtSetArg(args[n], XtNlabel,         text); n++;

   w = XtCreateManagedWidget(name,
                             labelWidgetClass,
                             parent, args, n);
   return w;
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       Widget createText
 *
 *  FUNCTION:   create a text widget
 *
 *  INPUT:      Widget       parent
 *              char        *name
 *              char        *text
 *              int          editable
 *              int          caret     show caret
 *              XFontStruct *font
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

Widget createText(Widget parent, char *name, char *text, int editable, int caret, XFontStruct *font, Widget top)
{
   int    n;
   Arg    args[20];
   Widget w;
   
   n = 0;
   if ( text )
   {
      XtSetArg(args[n],XtNstring, text);
      n++;
   }

   XtSetArg(args[n], XtNeditType,  caret?XawtextEdit:XawtextRead); n++;

   if ( editable )
   {
      XtSetArg(args[n], XtNscrollVertical, XawtextScrollAlways); n++;
      XtSetArg(args[n], XtNscrollHorizontal, XawtextScrollAlways); n++;
   }

   if ( font )
   {
      XtSetArg(args[n],XtNfont, font); n++;
   }

   XtSetArg(args[n], XtNtop,      XtChainTop);n++;
   XtSetArg(args[n], XtNbottom,   XtChainTop);n++;
   XtSetArg(args[n], XtNfromVert, top);n++;
   XtSetArg(args[n], XtNleft,     XtChainLeft);n++;
   XtSetArg(args[n], XtNright,    XtChainLeft);n++;

   w = XtCreateManagedWidget(name,
                             asciiTextWidgetClass,
                             parent, args, n);
   XawTextDisplayCaret(w, caret);
   return w;
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       Widget createRText
 *
 *  FUNCTION:   create a text widget
 *
 *  INPUT:      Widget       parent
 *              char        *name
 *              char        *text
 *              int          editable
 *              int          caret     show caret
 *              XFontStruct *font
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

Widget createRText(Widget parent, char *name, Widget top, Widget left, int caret)
{
   int    n;
   Arg    args[20];
   Widget w;
   
   n = 0;
   XtSetArg(args[n], XtNeditType, XawtextEdit); n++;
   XtSetArg(args[n], XtNtop,      XtChainTop);n++;
   XtSetArg(args[n], XtNbottom,   XtChainTop);n++;
   XtSetArg(args[n], XtNfromVert, top);n++;
   XtSetArg(args[n], XtNleft,     XtChainLeft);n++;
   XtSetArg(args[n], XtNright,    XtChainLeft);n++;
   XtSetArg(args[n], XtNfromHoriz, left);n++;

   w = XtCreateManagedWidget(name,
                             asciiTextWidgetClass,
                             parent, args, n);
   XawTextDisplayCaret(w, caret);
   return w;
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       Widget createButton
 *
 *  FUNCTION:   create a text widget
 *
 *  INPUT:      Widget       parent
 *              char        *name
 *              char        *label
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

Widget createButton(Widget parent, char *name, char *label, XtPointer cb, XtPointer val, Widget top, Widget left)
{
   int    n;
   Arg    args[10];
   Widget w;
   
   n = 0;
   if ( label )
   {
      XtSetArg(args[n],XtNlabel, label);
      n++;
   }
   XtSetArg(args[n],XtNtop,       XtChainTop);n++;
   XtSetArg(args[n],XtNbottom,    XtChainTop);n++;
   XtSetArg(args[n],XtNleft,      XtChainLeft);n++;
   XtSetArg(args[n],XtNright,     XtChainLeft);n++;
   XtSetArg(args[n],XtNfromVert,  top); n++;
   XtSetArg(args[n],XtNfromHoriz, left); n++;
   w = XtCreateManagedWidget(name,
                                  commandWidgetClass,
                                  parent, args, n);
   if ( cb )
   {
      XtAddCallback( w, XtNcallback, cb, (XtPointer)val);
   }
   return w;
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       Widget createTitleLabel
 *
 *  FUNCTION:   create a text widget
 *
 *  INPUT:      Widget       parent
 *              char        *name
 *              char        *title
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

Widget createTitleLabel(Widget parent, char *name, char *title)
{
   int    n;
   Arg    args[10];
   Widget w;
   
   n = 0;
   XtSetArg(args[n], XtNborderWidth,   0);n++;
   XtSetArg(args[n], XtNinternalWidth, 0);n++;
   XtSetArg(args[n], XtNlabel,         title); n++;
   XtSetArg(args[n], XtNbottom,        XtChainTop); n++;
   XtSetArg(args[n], XtNtop,           XtChainTop); n++;
   XtSetArg(args[n], XtNleft,          XtChainLeft);n++;
   XtSetArg(args[n], XtNright,         XtChainLeft);n++;
   w = XtCreateManagedWidget(name,
                             labelWidgetClass,
                             parent, args, n);
   return w;
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void setShellGeoAndTitle
 *
 *  FUNCTION:   set the size and position of the shell
 *
 *  INPUT:      Widget     shell
 *              Dimension  w       width of window
 *              Dimension  h       height of window
 *              char      *title
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void setShellGeoAndTitle(Widget shell, Dimension w, Dimension h, char *title)
{
   int n = 0;
   Arg args[20];
   Position x = (screenW - w)/2; /* preset X position */
   Position y = (screenH - h)/2; /* preset Y position */

   if ( strstr(appResourceRec.location, "top") || strstr(appResourceRec.location, "Top") )
   {
      y = appResourceRec.topMargin;
   }
   else if ( strstr(appResourceRec.location, "bottom") || strstr(appResourceRec.location, "Bottom") )
   {
      y = screenH - h - appResourceRec.bottomMargin;
   }
   
   if ( strstr(appResourceRec.location, "left") || strstr(appResourceRec.location, "Left") )
   {
      x = appResourceRec.leftMargin;
   }
   else if ( strstr(appResourceRec.location, "right") || strstr(appResourceRec.location, "Right") )
   {
      x = screenW - w - appResourceRec.rigthMargin;
   }

   if ( geometry )
   {
      XtSetArg(args[n],XtNgeometry, geometry);n++;
   }
   else if (appResourceRec.center )
   {
      if ( h + appResourceRec.topMargin + appResourceRec.bottomMargin >= screenH )
      {
         y = appResourceRec.topMargin;
      }
      if ( w + appResourceRec.rigthMargin + appResourceRec.leftMargin >= screenW )
      {
         x = appResourceRec.rigthMargin;
      }
      XtSetArg(args[n],XtNx,x);n++;
      XtSetArg(args[n],XtNy,y);n++;
   }
   else if ( appResourceRec.location )
   {
      /* if both values are 0 this will not work
       * we set them to 1
       */
      if ( x == 0 ) x = 1;
      if ( y == 0 ) y = 1;
      XtSetArg(args[n],XtNx,x);n++;
      XtSetArg(args[n],XtNy,y);n++;
   }

   if ( title )
   {
       XtSetArg(args[n],XtNtitle, title);n++;
   }

   XtSetArg(args[n], XtNminHeight, h);n++;
   XtSetArg(args[n], XtNmaxHeight, h);n++;
   XtSetArg(args[n], XtNminWidth,  w);n++;
   XtSetArg(args[n], XtNmaxWidth,  w);n++;
   XtSetArg(args[n], XtNmappedWhenManaged, True);n++;
   XtSetArg(args[n], XtNheight, h);n++;
   XtSetArg(args[n], XtNwidth, w);n++;
   XtSetArg(args[n], XtNmappedWhenManaged, True);n++;
   XtSetValues(shell, args,n);


}

#if 0
/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void setFocusToButton
 *
 *  FUNCTION:   set the focus to the passed button
 *
 *  INPUT:      Widget w
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void setFocusToButton(Widget w)
{
   static XEvent ev;

   XtVaSetValues(w, XtNhighlightThickness, 2, NULL);
   memset(&ev, 0,sizeof(ev));
   ev.xcrossing.type = EnterNotify;
   ev.xcrossing.display = XtDisplay(w);
   ev.xcrossing.window  = XtWindow(w);
   ev.xcrossing.subwindow  = 0;
   ev.xcrossing.send_event = True;
   ev.xcrossing.x = 1;
   ev.xcrossing.y = 1;
   ev.xcrossing.x_root = 0;
   ev.xcrossing.y_root = 0;
   ev.xcrossing.mode = NotifyNormal;
   XSendEvent(XtDisplay(w), XtWindow(w), True, EnterWindowMask, &ev);
}
#endif

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void addChoice
 *
 *  FUNCTION:   fill the struct for the presented choices
 *
 *  INPUT:      char *argv[]
 *
 *  REMARKS:    *argv[] must contain 2 values
 *
 *  +++fhe
 *********************************************************************
*/

void addChoice(char *argv[])
{

   if ( choice.count == 0 )
   {
      choice.text = (char**)malloc(sizeof(char**)*2);
      choice.ret  = (char**)malloc(sizeof(char**)*2);
   }
   else
   {
      choice.text = (char**)realloc(choice.text,
                                    sizeof(char**) * (choice.count+2));
      choice.ret  = (char**)realloc(choice.ret ,
                                    sizeof(char**) * (choice.count+2));

   }

   if ( choice.text == NULL || choice.ret == NULL )
   {
      fprintf(stderr,"%s: can't alloc memory for choices\n", program);
      exit(1);
   }

   choice.text[choice.count] = argv[0];
   if ( argv[1][0] == '\0' )
   {
      choice.ret[choice.count] = argv[0];
   }
   else
   {
      choice.ret[choice.count] = argv[1];
   }
   choice.count++;
   choice.text[choice.count] = NULL;
   choice.ret[choice.count]  = NULL;
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void addCheck
 *
 *  FUNCTION:   fill the struct for the presented checks
 *
 *  INPUT:      char *argv[]
 *
 *  REMARKS:    *argv[] must contain 2 values
 *
 *  +++fhe
 *********************************************************************
*/

void addCheck(char *argv[])
{
   check_t *n;
;
   if ( check == NULL )
   {
      check = (check_t*)calloc(sizeof(check_t),1);
      n = check;
   }
   else
   {
      n = check;
      while ( n->next )
         n = n->next;
      n->next = (check_t*)calloc(sizeof(check_t),1);
      n = n->next;
   }

   if ( n == NULL )
   {
      fprintf(stderr,"%s: can't alloc memory for items\n", program);
      exit(1);
   }

   n->text = argv[0];
   if ( argv[1][0] == '\0' )
   {
      n->ret = argv[0];
   }
   else
   {
      n->ret = argv[1];
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void cancelCB
 *
 *  FUNCTION:   print cancel string and return
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void cancelCB(Widget widget, XtPointer clientData, XtPointer callData)
{
   if ( cancelRet && *cancelRet )
      printf("%s\n",cancelRet);
   exit(3);
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void invisibleInput_CB
 *
 *  FUNCTION:   input for secret input fields
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void invisibleInput_CB(Widget w, XtPointer closure, XEvent *event, Boolean *contDisp)
{
   KeySym           key = 0;
   Modifiers       *modifiers = NULL;
   int              len;
   XawTextPosition  pos;

   *contDisp = True;

   if (event->xkey.type == KeyPress )
   {
      pos = XawTextGetInsertionPoint(w);
      key = XtGetActionKeysym(event, modifiers);

      /* replace xkeycode with the code for '*' */
      if ( key == XK_Tab || key == XK_ISO_Left_Tab )
      {
         *contDisp = False;
         return;
      }

      if ( key <= XK_asciitilde && key > XK_space)
      {
         event->xkey.keycode = XKeysymToKeycode(XtDisplay(w), XK_asterisk);
         if ( XtGetActionKeysym(event, modifiers) != XK_asterisk)
         {
            event->xkey.state = ShiftMask;
            event->xkey.keycode = XKeysymToKeycode(XtDisplay(w),XK_asterisk);
         }

         if ( passwdInput == NULL )
         {
            passwdInput = (char*)calloc(2,1);
            len = 0;
         }
         else
         {
            len = strlen(passwdInput);
            passwdInput = (char *)realloc(passwdInput, len + 2);
         }

         /* insert the new char at the right place */
         if ( pos == len )
         {
            passwdInput[len] = (char)key;
         }
         else
         {
            char new = (char)key;
            while ( pos <= len )
            {
               char old = passwdInput[pos];
               passwdInput[pos] = new;
               new = old;
               pos++;
            }
         }
         len++;
         passwdInput[len] = '\0';
      }
      else
      {
         if ( passwdInput != NULL )
         {
            len = strlen(passwdInput);
         }
         switch(key)
         {
            case XK_BackSpace:
               /* remove character left of position */
               if ( passwdInput )
               {
                  len = strlen(passwdInput);
                  if ( pos > 0 )
                  {
                     strcpy(&passwdInput[pos-1], &passwdInput[pos]);
                     len--;
                     if ( len >= 0 )
                     {
                        passwdInput[len] = '\0';
                     }
                  }
               }
            break;
            case XK_Delete:
               /* remove character right of position */
               if ( passwdInput )
               {
                  len = strlen(passwdInput);
                  if ( len > pos )
                  {
                     strcpy(&passwdInput[pos], &passwdInput[pos+1]);
                     len--;
                     if ( len >= 0 )
                     {
                        passwdInput[len] = '\0';
                     }
                  }
               }
            break;
            case XK_Left:
                  if ( pos > 0 )
                  {
                     pos--;
                  }
            break;
            case XK_Right:
               if ( passwdInput != 0 )
                  len = strlen(passwdInput);
               else
                  len = 0;
               if ( pos < len )
               {
                  pos++;
               }
            break;
            default:
               *contDisp = False;
         }
      }
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void do_selection_callback
 *
 *  FUNCTION:   handle paste into passwd widget
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void do_selection_callback(Widget        w,
                          XtPointer      clientData,
                          Atom          *selection,
                          Atom          *type,
                          caddr_t        value,
                          unsigned long *length,
                          int           *format)
{
   int        i;
   XEvent     event;
   Modifiers *modifiers = NULL;
   long       len = 0;

   len = (int)*length;
   /* checks */
   if ( *type == XT_CONVERT_FAIL )
   {
      XBell( XtDisplay(w), 100 );
      return;
   }

   if ( *length == 0 || value == NULL )
   {
      return;
   }

   if ( *selection != XA_PRIMARY )
   {
      len = 0;
   }

   if ( *format != 8 )
   {
      len = 0;
   }

   if ( *type != XA_STRING )
   {
      len = 0;
   }

   len = (int)*length;
   for ( i = 0; i < len; i++ )
   {
      if ( isspace(value[i]) || iscntrl(value[i]) )
      {
         len = 0;
         break;
      }
      if ( ! isprint(value[i]) || ! isascii(value[i]) )
      {
         len = 0;
         break;
      }
   }
   if ( len == 0 )
   {
      XBell( XtDisplay(w), 100 );
   }

   /* insert into the passwd widget */
   for ( i = 0; i < len ; i++ )
   {
      memset(&event, 0, sizeof(event));
      event.xkey.type       = KeyPress;
      event.xkey.display    = XtDisplay(passwdIn);
      event.xkey.window     = XtWindow(passwdIn);
      event.xkey.root       = XtWindow(toplevel);
      event.xkey.subwindow  = XtWindow(passwdIn);
      event.xkey.send_event = True;
      event.xkey.time       = CurrentTime;
      event.xkey.keycode    = XKeysymToKeycode(XtDisplay(passwdIn), value[i]);
      if ( XtGetActionKeysym(&event, modifiers) != value[i])
      {
         event.xkey.state = ShiftMask;
         event.xkey.keycode = XKeysymToKeycode(XtDisplay(passwdIn),value[i]);
      }
      XSendEvent(event.xkey.display,
                 event.xkey.window, True,
                 KeyPressMask, &event);
   }

   XtFree( value );
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void endHelpCB
 *
 *  FUNCTION:   unmap help window
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void endHelpCB(Widget widget, XtPointer clientData, XtPointer callData)
{
   XtUnmapWidget(hlplevel);
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void HelpCB
 *
 *  FUNCTION:   build help window or map it if allready created
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void helpCB(Widget widget, XtPointer clientData, XtPointer callData)
{
   Arg    args[100];
   int    n;
   char  *buf = NULL;
   Dimension w1 = 0, w2 = 0;
   Dimension iw = 0;
   Dimension h1 = 0, h2 = 0;
   Dimension bw1, bw2;
   Dimension wMax, hMax;
   Dimension dist = 0;
   Dimension hSpace = 4;
   Dimension vSpace = 4;
   Widget wid;
   int lines   = 0;
   int columns = 0;

   if ( hlplevel )
   {
      XtMapWidget(hlplevel);
      return;
   }

   getText(helpFile, &buf);

   /* shell */
   n = 0;
   XtSetArg(args[n],XtNmappedWhenManaged, False);n++;
   hlplevel = XtAppCreateShell(class, "Help",
                               transientShellWidgetClass,
                               XtDisplay(toplevel),
                               args,
                               n);
   /* form */
   helpW = createContainer(hlplevel, "helpForm");

   /* Text */
   helpM = createText(helpW, "text", buf, False, False, NULL, NULL); 

   /* Button */
   wid = createButton(helpW, "okButton", helpOK, endHelpCB,  NULL, helpM, NULL);
   XtRealizeWidget(hlplevel);

   getSpacing(helpW, &hSpace, &vSpace);

   /* get sizes */
   lines=0;
   calcTextSize(helpM, buf, &lines, &columns, &h1, &w1);

   /* if the size is to big, limit them */
   if ( w1 > screenW - appResourceRec.topMargin - appResourceRec.bottomMargin - hSpace - hSpace )
   {
      w1 = screenW - appResourceRec.topMargin - appResourceRec.bottomMargin - hSpace - hSpace;
      XtVaSetValues(helpM,
                    XtNscrollHorizontal, XawtextScrollAlways,
                    NULL);
   }

   if ( h1 > screenH - appResourceRec.topMargin - appResourceRec.bottomMargin - hSpace - hSpace)
   {
      h1 = screenH - appResourceRec.topMargin - appResourceRec.bottomMargin - hSpace - hSpace;
      XtVaSetValues(helpM,
                    XtNscrollVertical, XawtextScrollAlways,
                    NULL);
      XtAddEventHandler(helpM,
                        KeyPressMask|KeyReleaseMask|ButtonReleaseMask|ButtonPressMask,
                        True,
                        (XtEventHandler) scrollMsg_CB, (XtPointer)True);
   }
   else
   {
      XtAddEventHandler(helpM,
                        KeyPressMask|KeyReleaseMask|ButtonReleaseMask|ButtonPressMask,
                        True,
                        (XtEventHandler) scrollMsg_CB, (XtPointer)False);
   }
   
   XtVaGetValues(helpM,
                 XtNborderWidth, &bw1,
                 XtNleftMargin,  &iw,
                 NULL);
   w1 += iw * 2; /* right margin is to be the same a the left margin */

   XtVaGetValues(wid,
                 XtNwidth,       &w2,
                 XtNheight,      &h2,
                 XtNborderWidth, &bw2,
                 NULL);

   wMax = bw1 * 2 + w1;

   if ( w2 + bw2 * 2 > wMax )
   {
      wMax = w2 + bw2 * 2;
   }

   if ( w2 < appResourceRec.minButtonWidth )
   {
      w2 = appResourceRec.minButtonWidth;
   }

   setWidthAndHeight(helpM,  wMax - bw1*2, h1);

   /* width of spcacer widget */
   switch(*appResourceRec.buttonAlignment)
   {
      case 'R':
         dist = wMax + hSpace - (bw2 + w2 + bw2);
      break;
      case 'L':
         dist = hSpace;
      break;
      default:
         dist = hSpace + (wMax - (bw2 + w2 + bw2))/2;
   }

   /* button */
   setWidthAndX(wid, w2, dist);

   /* we have to resize the shell so we set the min and max */
   /* width and height                                      */

   hMax = vSpace + bw1 + h1 + bw1 + vSpace + bw2 + h2 + bw2 + vSpace;
   setShellGeoAndTitle(hlplevel, hSpace+wMax+hSpace, hMax, title);
   XtMapWidget(hlplevel);
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void endErrorCB
 *
 *  FUNCTION:   exit
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void endErrorCB(Widget widget, XtPointer clientData, XtPointer callData)
{
   exit(0);
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void scrollMsg_CB
 *
 *  FUNCTION:   scrolling within the non editable text widget 
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void scrollMsg_CB(Widget w, XtPointer closure, XEvent *event, Boolean *contDisp)
{
   KeySym     key;
   Modifiers *modifiers = NULL;
   int        allowScroll = (int)closure;

   if ( allowScroll )
   {
      if ( event->xany.type == KeyPress )
      {
         key = XtGetActionKeysym(event, modifiers);
         switch(key)
         {
            case XK_Page_Up:
            case XK_KP_Page_Up:
               XtCallActionProc(w, "previous-page", event, NULL, 0);
               break;
            case XK_Page_Down:
            case XK_KP_Page_Down:
               XtCallActionProc(w, "next-page", event, NULL, 0);
               break;
            break;
            case XK_Home:
            case XK_KP_Home:
               XtCallActionProc(w, "beginning-of-file", event, NULL, 0);
               break;
            case XK_End:
            case XK_KP_End:
               XtCallActionProc(w, "end-of-file", event, NULL, 0);
               break;
            case XK_Return:
            case XK_KP_Enter:
            case XK_space:
               /* == press of  OK button Callback */
               if ( w == msgT || w == inText  )
                  exit(0);
               else
                  XtUnmapWidget(XtParent(XtParent(w)));
               break;
            default:
               *contDisp = False;
         }
      }
      else if ( event->xany.type == ButtonPress )
      {
         switch ( event->xbutton.button )
         {
            case 4:
               XtCallActionProc(w, "previous-page", event, NULL, 0);
               *contDisp = False;
            break;
            case 5:
               XtCallActionProc(w, "next-page", event, NULL, 0);
               *contDisp = False;
            break;
            default: *contDisp = True;
         }
      }
      else if ( event->xany.type == ButtonRelease )
      {
         switch ( event->xbutton.button )
         {
            case 4:
            case 5:
               *contDisp = False;
            break;
            default:
               *contDisp = True;
         }
      }
   }
   else
   {
      if ( event->xany.type == KeyPress )
      {
         key = XtGetActionKeysym(event, modifiers);
         switch(key)
         {
            case XK_Return:
            case XK_KP_Enter:
            case XK_space:
                /* == press of  OK button Callback */
                if ( w == msgT || w == inText )
                   exit(0);
                else
                   XtUnmapWidget(XtParent(XtParent(w)));
            default:
               *contDisp = False;
         }
      }
#ifdef XAW3D
      else
      {
         /* workaround for a bug within Xaw3d ? */
         if ( event->xany.type == ButtonPress
              ||
              event->xany.type == ButtonRelease
              ||
              event->xany.type == MotionNotify
            )
         {
            *contDisp = False;
            
         }
      }
#endif
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void calcTextSize
 *
 *  FUNCTION:   get width and heigth in pixel for the given text
 *              widget
 *
 *  INPUT:      Widget w
 *              String text
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

int calcTextSize(Widget wid, char *text, int *lines, int *columns, Dimension *height, Dimension *width)
{
   XFontStruct *font;
   Dimension w;
   char *s, *t;

   *height  = 0;
   *width   = 0;

   /* get font */
   XtVaGetValues(wid,
                 XtNfont, &font,
                 NULL);

   if ( font == NULL )
   {
      return False;
   }

   t = s = text;
   while ( s && *s )
   {
      s = strchr(s, '\n');
      if ( s )
      {
         (*lines)++;
         w = XTextWidth(font, t, s-t);
         *width = w > *width ? w : *width;
         *columns = *columns > (s-t) ? *columns : (s-t);
         s++;
         t = s;
      }
      else
      {
         /* line not terminated with '\n' */
         *columns = strlen(t);
         w = XTextWidth(font, t,*columns);
         (*lines)++;
         *width = w > *width ? w : *width;
      }
   }

   /* now the height */
   *height = (font->max_bounds.ascent +
              font->max_bounds.descent) *  *lines;

   return True;
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void showError
 *
 *  FUNCTION:   exit
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void showError()
{
   char       *buf;
   int          n;
   Dimension    w1 = 0, w2 = 0, h1 = 0, h2 = 0, bw2 = 0;
   Dimension    maxW, maxH;
   Dimension    distW;
   Dimension    hSpace  = 4;
   Dimension    vSpace  = 4;
   Dimension    tbw     = 0;
   Dimension    lm      = 0;
   int          lines   = 1;
   int          columns = 0;
   XFontStruct *fontn = NULL;
   XrmValue     fromVal, toVal;
   Widget       wid;
   int          scrollFlag = False;

   if ( errFile )
   {
      getText(errFile, &buf);

      /******************************************************/
      /* get fonts                                          */
      /******************************************************/

      if ( fn )
      {
         fromVal.addr = fn;
         toVal.addr = NULL;
         fromVal.size = sizeof(String);
         XtConvert(toplevel, XtRString, (XrmValuePtr) &fromVal, XtRFontStruct,
                   (XrmValuePtr) &toVal);
         if ( toVal.addr != NULL )
         {
            fontn = (XFontStruct*) *((XFontStruct **)toVal.addr);
         }
      }

      /* top form */
      helpW = createContainer(toplevel, error ? "error" : "message");
      getSpacing(toplevel, &hSpace, &vSpace);

      /* label, really a text widget */
      msgT = createText(helpW, "text", buf, False, False, fontn, NULL);

      /* close button */
      wid = createButton(helpW, "okButton",
                         error ? error :
                         message ? message : Wait, endErrorCB, NULL, msgT, NULL);
      XtOverrideTranslations(wid,
                             XtParseTranslationTable(buttonTranslation));

      XtRealizeWidget(toplevel);
      getSpacing(helpW, &hSpace, &vSpace);

      /* size for message */
      lines   = 0;
      columns = strlen(error?error:message?message:Wait);
      calcTextSize(msgT, buf, &lines, &columns, &h1, &w1);
      XtVaGetValues(msgT,
                    XtNborderWidth, &tbw,
                    NULL);

      XtVaGetValues(msgT, XtNleftMargin, &lm, NULL);
      h1 += lm * 2;
      maxW = w1 + tbw + tbw + lm * 2;

      if ( maxW > screenW - appResourceRec.leftMargin - appResourceRec.rigthMargin )
      {
         maxW = screenW - appResourceRec.leftMargin - appResourceRec.rigthMargin;
         w1 = maxW - tbw - tbw - 4;
         XtVaSetValues(msgT,
                       XtNscrollHorizontal, XawtextScrollAlways,
                       NULL);
         h1 += 13; 
         scrollFlag = True;
      }


      maxH = h1 + tbw + tbw;
      if ( maxH > screenH - appResourceRec.topMargin - appResourceRec.bottomMargin )
      {
         maxH = screenH - appResourceRec.topMargin - appResourceRec.bottomMargin;
         h1 = maxH;
         maxW += SCRBW; /* + size of scrollbar */
         XtVaSetValues(msgT,
                       XtNscrollVertical, XawtextScrollAlways,
                       NULL);
         scrollFlag = True;
      }
      XtAddEventHandler(msgT,
                        KeyPressMask|KeyReleaseMask|ButtonReleaseMask|ButtonPressMask,
                        True,
                        (XtEventHandler) scrollMsg_CB, (XtPointer)scrollFlag);

      /* size for button */
      XtVaGetValues(wid,
                    XtNwidth,       &w2,
                    XtNheight,      &h2,
                    XtNborderWidth, &bw2,
                    NULL);

      if ( w2 < appResourceRec.minButtonWidth )
      {
         w2 = appResourceRec.minButtonWidth;
      }

      if ( w2 > maxW )
      {
         n =  bw2 + w2 + bw2;
         maxW = n > maxW ? n : maxW;
      }
      maxH = maxH + h2 + 2 * bw2;

      /* set new width and height */
      setWidthAndHeight(msgT, maxW-tbw-tbw, h1);

      /* space */
      switch(*appResourceRec.buttonAlignment)
      {
         case 'R':
            distW = (maxW - bw2 - w2 - bw2)+ hSpace;
         break;
         case 'L':
             distW = hSpace;
         break;
         default:
           distW = (maxW - bw2 - w2 - bw2)/2+hSpace;
      }
      XtVaSetValues(wid, XtNhorizDistance, distW,NULL);
      setWidth(wid, w2);

      maxH += vSpace;
      maxW += hSpace;

      setShellGeoAndTitle(toplevel, maxW+hSpace, maxH+vSpace*2, title);
      XtMapWidget(toplevel);
      XtSetKeyboardFocus( toplevel, msgT);
      XtCallAcceptFocus(msgT,CurrentTime);
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void setFocus_CB
 *
 *  FUNCTION:   select widget with Focus onm TAB SHIFT+TAB
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void setFocus_CB(Widget w, XtPointer closure, XEvent *event, Boolean *contDisp)
{
   KeySym     key;
   Modifiers *modifiers=NULL;
   Boolean    res;
   widTab_t  *wTab;

   *contDisp = True;
   if ( event->xany.type == KeyPress || event->xany.type == KeyRelease )
   {
      key = XtGetActionKeysym(event, modifiers);

      /* check wich widget will be the next*/
      if ( (key == XK_Tab || key == XK_ISO_Left_Tab ) &&
           event->xkey.type == KeyPress )
      {
         wTab = widgetTab;
         while ( wTab )
         {
            if ( wTab->wid == w )
            {
               if ( key == XK_Tab )
               {
                  if ( wTab->next == NULL )
                  {
                     wTab = widgetTab;
                  }
                  else
                  {
                     wTab = wTab->next;
                  }
               }
               else
               {
                  if ( wTab->prev != NULL )
                  {
                     wTab = wTab->prev;
                  }
                  else
                  {
                     /* set last element */
                     while ( wTab->next )
                        wTab = wTab->next;
                  }
               }
               break;
            }
            wTab = wTab->next;
         }
      }
      else
      {
         /* handle Return within list or input text field */
         if ( event->xany.type == KeyPress &&
             (key == XK_Return || key == XK_KP_Enter || key == XK_space || key == XK_KP_Space) )
         {
            if ( w == list )
            {
               okCB(w, NULL, (XtPointer)XWCHOICE_KB);
            }
            else if ( w == inText && (key == XK_Return || key == XK_KP_Enter)
                )
            {
               okCB(w, NULL, (XtPointer)XWCHOICE_KB);
            }
            else if ( (w == passwdIn || w == loginIn )
                    && (key == XK_Return || key == XK_KP_Enter)
                )
            {
               passwdCB(w, NULL, NULL);
            }
            else if (strcmp(XtName(w), "check") == 0)
            {
               /* toggle */
               if ( radio )
               {
                  toggleCB(w, NULL, NULL);
               }
               else
               {
                  short state;
                  XtVaGetValues(w,XtNstate,&state, NULL);
                  state = state ? 0 : 1;
                  XtVaSetValues(w,XtNstate,state, NULL);
               }
            }
         }
         return;
      }
   }
   else if ( event->xany.type == ButtonRelease ||
            (event->xany.type == ButtonPress && w == inText )
           )
   {
      wTab = widgetTab;
      while ( wTab )
      {
         if ( wTab->wid == w )
         {
            XtSetKeyboardFocus( toplevel, wTab->wid);
            res = XtCallAcceptFocus(wTab->wid,CurrentTime);
            break;
         }
         wTab = wTab->next;
      }
      if ( wTab == NULL )
      {
         return;
      }
   }
   else if ( event->xany.type == ButtonRelease ||
            (event->xany.type == ButtonPress && (w == passwdIn || w == loginIn)  )
           )
   {
      wTab = widgetTab;
      while ( wTab )
      {
         if ( wTab->wid == w )
         {
            XtSetKeyboardFocus( toplevel, wTab->wid);
            res = XtCallAcceptFocus(wTab->wid,CurrentTime);
            break;
         }
         wTab = wTab->next;
      }
      if ( wTab == NULL )
      {
         return;
      }
   }
   else if ( event->xany.type == ConfigureNotify )
   {
      wTab = widgetTab;
      while ( wTab )
      {
         if ( wTab->wid == list || wTab->wid == vp)
         {
            Position y;
            if ( vp )
            {
               XtVaSetValues(vp, XtNborderColor, darkblue, NULL);
            }
            else
            {
               XtVaSetValues(list, XtNborderColor, darkblue, NULL);
            }
            selectedWidget =  list;
            XtSetKeyboardFocus( toplevel, list);
            res = XtCallAcceptFocus(list,CurrentTime);
            XtVaGetValues(list,XtNy,&y,NULL);
            return;
         }
         
         wTab = wTab->next;
      }
   }
   else
   {
      return;
   }

   if (wTab)
   {
      XtSetKeyboardFocus( toplevel, wTab->wid);
      res = XtCallAcceptFocus(wTab->wid,CurrentTime);

      if ( wTab->wid == list||wTab->wid == vp)
      {
         if ( vp )
         {
            XtVaSetValues(vp, XtNborderColor, darkblue, NULL);
         }
         else
         {
            XtVaSetValues(list, XtNborderColor, darkblue, NULL);
         }
      }
      else if ( list != NULL  )
      {
         if ( vp )
         {
            XtVaSetValues(vp, XtNborderColor, dgrey, NULL);
         }
         else
         {
            XtVaSetValues(list, XtNborderColor, dgrey, NULL);
         }
      }

      if ( wTab->wid == inText )
      {
         XtVaSetValues(inText, XtNborderColor, darkblue, NULL);
      }
      else if ( inText != NULL )
      {
         XtVaSetValues(inText, XtNborderColor, dgrey, NULL);
      }

      if ( wTab->wid == passwdIn )
      {
         XtVaSetValues(passwdIn, XtNborderColor, darkblue, NULL);
         XawTextDisplayCaret(passwdIn, True);
      }
      else if ( passwdIn != NULL )
      {
         XtVaSetValues(passwdIn, XtNborderColor, dgrey, NULL);
         XawTextDisplayCaret(passwdIn, False);
     }

      if ( wTab->wid == loginIn )
      {
         XtVaSetValues(loginIn, XtNborderColor, darkblue, NULL);
         XawTextDisplayCaret(loginIn, True);
      }
      else if ( loginIn != NULL )
      {
         XtVaSetValues(loginIn, XtNborderColor, dgrey, NULL);
          XawTextDisplayCaret(loginIn, False);
      }

      if ( wTab->wid == cancelB )
      {
         XtVaSetValues(cancelB, XtNhighlightThickness, 2, NULL);
      }
      else if ( cancelB != NULL )
      {
         XtVaSetValues(cancelB, XtNhighlightThickness, 0, NULL);
      }

      if ( wTab->wid == helpB )
      {
         XtVaSetValues(helpB, XtNhighlightThickness, 2, NULL);
      }
      else if ( helpB != NULL )
      {
         XtVaSetValues(helpB, XtNhighlightThickness, 0, NULL);
      }

      if ( wTab->wid == okB )
      {
         XtVaSetValues(okB, XtNhighlightThickness, 2, NULL);
      }
      else if ( okB != NULL )
      {
         XtVaSetValues(okB, XtNhighlightThickness, 0, NULL);
      }

      if ( strcmp(XtName(wTab->wid),"check") == 0 )
      {
         XtVaSetValues(wTab->wid, XtNborderColor, darkblue, NULL);
      }

      if ( strcmp(XtName(w),"check") == 0 )
      {
         XtVaSetValues(w, XtNborderColor, dgrey, NULL);
      }

      selectedWidget =  wTab->wid;
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void leave_CB
 *
 *  FUNCTION:   handle enter leave of pushbutton
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/
void leave_CB(Widget w, XtPointer closure, XEvent *event, Boolean *contDisp)
{
   int isButton = selectedWidget &&
                  (selectedWidget == okB ||
                   selectedWidget == cancelB ||
                   selectedWidget == helpB ||
                   (strcmp(XtName(selectedWidget),"check") == 0) );
   if ( event->xcrossing.type == EnterNotify )
   {
      *contDisp = True;
      if ( w != selectedWidget && selectedWidget && isButton )
      {
         XtVaSetValues(selectedWidget, XtNhighlightThickness, 0, NULL);
      }
   }
   else
   {
      *contDisp = True;
      if ( w == selectedWidget )
      {
         XtVaSetValues(w, XtNhighlightThickness, 2, NULL);
         *contDisp = False;
      }
      else
      {
         XtVaSetValues(w, XtNhighlightThickness, 0, NULL);
      }

      /* selected is a button set Highlight */
      if ( isButton )
      {
         XtVaSetValues(selectedWidget, XtNhighlightThickness, 2, NULL);
      }
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void dblKlick
 *
 *  FUNCTION:   reset the variable which tell us that the
 *              timer delay for double klick has not elapsed
 *
 *  INPUT:      XtPointer closure pointer to the variable
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void dblKlick(XtPointer closure, XtInputId *id)
{
   int *val = (int*)closure;
   *val = 0;
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void setListPosition
 *
 *  FUNCTION:   select list item from keyboard cursor down and up keys
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void setListPosition(Widget w, int item, int inputType, int evType)
{
   Position    x=0;
   Position    y;
   Dimension   h;
   Dimension   wi, cspc, rspc, iw;
   Position    l,r;
   int         li, liTotal;
   int         cols = 1;
   int         column;
   Dimension   colSize;
   static  int lastItem = -1;
   static  int timeNotElapsed = 0;

   /* double click, check if slection ist comming from mouse */
   if ( inputType == XWCHOICE_MOUSE )
   {
       if ( timeNotElapsed && lastItem == item )
       {
          okCB(list, NULL, (XtPointer)XWCHOICE_MOUSE);
       }
       else
       {
          XtAppAddTimeOut(appContext,
                          appResourceRec.doubleKlickDelay,
                          dblKlick,
                          &timeNotElapsed);
          timeNotElapsed = 1;
       }
   }

   if ( columnsT )
     cols = atoi(columnsT);

   XtVaGetValues(list,
                 XtNheight,        &h,
                 XtNwidth,         &wi,
                 XtNlongest,       &colSize,
                 XtNcolumnSpacing, &cspc,
                 XtNrowSpacing,    &rspc,
                 XtNinternalWidth, &iw,
                 NULL);
   colSize += cspc ;

   if ( h > viewPortHeight || wi > viewPortWidth)
   {

      liTotal = (choice.count+(cols-1))/cols;
      li = item % liTotal;
      if ( li + 1 > lines )
      {
         y = ((float)(li) / (float)lines)*(viewPortHeight-4)-4;
      }
      else
      {
         y = 0;
      }

      column  = (item / liTotal);

      XtVaGetValues(list,
                    XtNx, &x,
                    NULL);
      x = -x;
      if ( colSize >= viewPortWidth )
      {
         x = colSize*(column);
      }
      else /* viewport width > column size */
      {
         l = colSize*column;
         r = colSize*(column+1);
         if ( l == 0 )
         {
            x = 0;
         }
         else  if ( r - x >= viewPortWidth )
         {
            x = r - viewPortWidth ;
         }
         else if ( l-x < 0 )
         {
            x = l - viewPortWidth ;
         }
      }
      if ( evType != ButtonRelease )
         XawViewportSetCoordinates(XtParent(w),x,y);
   }
   lastItem = item;
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void ListSel_CB
 *
 *  FUNCTION:   select list item from keyboard cursor down and up keys
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void ListSel_CB(Widget w, XtPointer closure, XEvent *event, Boolean *contDisp)
{
   KeySym                key;
   Modifiers             modifiers;
   XawListReturnStruct *selection;
   int                   newSelected;
   int                   cols = 1;
   int                   valid = False;
   int                   inputType = XWCHOICE_MOUSE; /* assume selection
                                                      * from the mouse */

   if ( event->xany.type == KeyPress )
   {
      Dimension x, y, spc;
      key = XtGetActionKeysym(event, &modifiers);
      XtVaGetValues(XtParent(list),
                    XtNheight, &spc,
                    NULL);
      XtVaGetValues(list,
                    XtNx, &x,
                    XtNy, &y,
                    NULL);

      x = - (x-2); /* internalHeight / Width previouslly set to 2 */
      y = - (y-2);
      spc -= 2;

      switch(key)
      {
         case XK_Next:
         case XK_KP_Next:
            spc = -spc;
         case XK_KP_Prior:
            y -= spc;
            XawViewportSetCoordinates(XtParent(w),x,y);
            return;
         case XK_Home:
         case XK_KP_Home:
            XawViewportSetCoordinates(XtParent(w),0,0);
            return;
         case XK_End:
         case XK_KP_End:
            XtVaGetValues(list, XtNheight,&y,NULL);
            XawViewportSetCoordinates(XtParent(w),x,y);
            return;
      }
   }

   if ( event->xany.type == ButtonPress || event->xany.type == ButtonRelease )
   {
      if ( event->xbutton.button > 3 )
      {
         if ( event->xany.type & ButtonRelease )
         {
            Dimension x, y, spc;
            XtVaGetValues(list,
                          XtNx, &x,
                          XtNy, &y,
                          XtNrowSpacing, &spc,
                          NULL);
            x = - (x-2);
            y = - (y-2);
            XtVaGetValues(XtParent(list),
                          XtNheight, &spc,
                          NULL);

            if ( event->xbutton.button == 4 )
               y -= spc/2;
            else if ( event->xbutton.button == 5 )
               y += spc/2;
            XawViewportSetCoordinates(XtParent(w),x,y);
            *contDisp = False;
            return;
         }
      }
   }

   if ( columnsT )
    cols = atoi(columnsT);

   *contDisp = True;
   cols = cols ? cols : 1;

   selection = XawListShowCurrent(list);

   if ( event->xkey.type == KeyRelease )
   {
      *contDisp = False;
      return;
   }

   /* set the event source to kbd if applicable */
   if ( event->xkey.type == KeyPress )
   {
      inputType = XWCHOICE_KB;
   }

   if ( selection == NULL )
   {
      setFocus_CB(w,closure,event,contDisp);
      return;
   }

   newSelected = selection->list_index;

   key = XtGetActionKeysym(event, &modifiers);

   switch(key)
   {
      case XK_Up:
      case XK_KP_Up:
         if ( selection->list_index > 0 )
         {
            newSelected = selection->list_index - 1;
            XawListHighlight(list, newSelected );
         }
      break;
      case XK_Down:
      case XK_KP_Down:
         if ( selection->list_index < choice.count -1 )
         {
            newSelected = selection->list_index + 1;
            XawListHighlight(list, newSelected );
         }
      break;
      case XK_Right:
      case XK_KP_Right:
      {
         int  cols = atoi(columnsT!=NULL?columnsT:"1");
         if ( cols > 1 )
         {
            int next;
            int delta = ((choice.count+cols-1)/cols);

            next = selection->list_index + delta ;
            if ( next > choice.count - 1 )
               next -= delta * cols;
            if ( next < 0 )
               next += delta;
            newSelected = next;
            XawListHighlight(list, next );
         }
      }
      break;
      case XK_Left:
      case XK_KP_Left:
      {
         int  cols = atoi(columnsT!=NULL?columnsT:"1");
         if ( cols > 1 )
         {
            int next;
            int delta = ((choice.count+cols-1)/cols);

            next = selection->list_index - delta ;
            if ( next < 0 )
               next += delta * cols;
            if ( next > choice.count - 1 )
               next -= delta;
            XawListHighlight(list, next );
            newSelected = next;
         }
      }
      break;
      case XK_Tab: case XK_ISO_Left_Tab:
         setFocus_CB(w,closure,event,contDisp);
      break;
      case XK_Return:
      case XK_KP_Enter:
      case XK_space:
      case XK_KP_Space:
         valid = True;
      break;
   }
   if ( event->xany.type != ButtonPress )
      setListPosition(w, newSelected, inputType, event->xany.type );

   if ( valid || inputType == XWCHOICE_MOUSE )
   {
      setFocus_CB(w,closure,event,contDisp);
   }

}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void okCB
 *
 *  FUNCTION:   return selection or beep in no selction
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void okCB(Widget widget, XtPointer clientData, XtPointer callData)
{
   XawListReturnStruct *selection;
   String               str = NULL;
   char                *s   = "";

   if ( noReturn )
   {
      /* return without printing any thinks */
      exit(0);
   }

   /***********************************************************/
   /* if we have a list and an input field the User data will */
   /* take precedence                                         */
   /***********************************************************/
   if ( input )
   {
      XtVaGetValues(inText,XtNstring, &str,NULL);

      if ( str )
      {
         s = str;
         while ( *s && isspace(*s)) s++;
         if (denyChars)
         {
            /* if any of the mentionned chars are in the inpt: deny */
            char *d = denyChars;
            while ( *d )
            {
               if ( strchr(s,*d) != NULL )
               {
                  XBell(XtDisplay(widget),100);
                  return;
               }
               d++;
            }
         }
      }

      if ( (str == NULL || *s == '\0') && (list == NULL || both))
      {
         XBell(XtDisplay(widget),100);
         return;
      }
      else if ( str && *s )
      {
         if ( !both )
         {
            printf("%s\n",s);
            exit(0);
         }
      }
   }

   /* look for list selection */
   if ( list )
   {
      selection = XawListShowCurrent(list);
      if (selection == NULL || selection->list_index< 0 )
      {
         XBell(XtDisplay(widget),100);
         return;
      }
      if ( !both )
      {
         printf("%s\n",choice.ret[selection->list_index]);
      }
      else
      {
         printf("%s %s\n",choice.ret[selection->list_index], s);
      }
   }
   exit(0);
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void passwdCB
 *
 *  FUNCTION:   return login and password according to
 *              configuration
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void passwdCB(Widget widget, XtPointer clientData, XtPointer callData)
{
   String   str = NULL;
   char    *s   = "";

   /***********************************************************/
   /* if we have a list and an input field the User data will */
   /* take precedence                                         */
   /***********************************************************/
   if ( loginTxt )
   {
      XtVaGetValues(loginIn,XtNstring, &str,NULL);
      if ( str )
      {
         s = str;
         while ( *s && isspace(*s)) s++;
      }

      if ( str && *s && passwdTxt == NULL)
      {
         printf("%s\n",s);
      }
      else
      {
         if ( str && passwdTxt )
         {
            printf("%s:%s\n", str, passwdInput ? passwdInput:"");
         }
      }
   }
   else
   {
      if (passwdInput)
      {
         printf("%s\n",passwdInput);
      }
      else
      {
         printf("\n");
      }
   }
   exit(0);
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void getColors
 *
 *  FUNCTION:   get a fews colors
 *              (if we have a display with 256 colors and
 *               can't get the wanted color we will default to
 *               black or white)
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void getColors()
{
   XrmValue     fromVal, toVal;

   fromVal.addr = "#0000cf";
   toVal.addr = NULL;
   fromVal.size = sizeof(String);
   XtConvert(toplevel, XtRString, &fromVal, XtRPixel, &toVal);
   if ( toVal.addr != NULL )
   {
      darkblue = *(Pixel*)toVal.addr;
   }
   else
   {
      fromVal.addr = "blue";
      toVal.addr = NULL;
      fromVal.size = sizeof(String);
      XtConvert(toplevel, XtRString, &fromVal, XtRPixel, &toVal);
      if ( toVal.addr != NULL )
      {
         darkblue = *(Pixel*)toVal.addr;
      }
      else
      {
         darkblue = BlackPixel(XtDisplay(toplevel),0);
      }
   }

   fromVal.addr = "#cfcfcf";
   toVal.addr = NULL;
   fromVal.size = sizeof(String);
   XtConvert(toplevel, XtRString, &fromVal, XtRPixel, &toVal);
   if ( toVal.addr != NULL )
      grey = *(Pixel*)toVal.addr;
   else
   {
      fromVal.addr = "#cccccc";
      toVal.addr = NULL;
      fromVal.size = sizeof(String);
      XtConvert(toplevel, XtRString, &fromVal, XtRPixel, &toVal);
      if ( toVal.addr != NULL )
         grey = *(Pixel*)toVal.addr;
      else
         grey = WhitePixel(XtDisplay(toplevel),0);
   }

   fromVal.addr = "#ff0000";
   toVal.addr = NULL;
   fromVal.size = sizeof(String);
   XtConvert(toplevel, XtRString, &fromVal, XtRPixel, &toVal);
   if ( toVal.addr != NULL )
      red = *(Pixel*)toVal.addr;
   else
   {
      red = BlackPixel(XtDisplay(toplevel),0);
   }

   fromVal.addr = "white";
   fromVal.size = sizeof(String);
   toVal.addr = NULL;
   XtConvert(toplevel, XtRString, &fromVal,   XtRPixel, &toVal);
   if ( toVal.addr != NULL )
      white = *(Pixel*)toVal.addr;
   else
   {
      white = WhitePixel(XtDisplay(toplevel),0);
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void buildLog
 *
 *  FUNCTION:   build the log window
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void buildLog()
{
   Dimension    wMax,w, hMax,h, bwText, bwButton, spc;
   Dimension    hSpace = 4;
   Dimension    vSpace = 4;
   XFontStruct *fontn=NULL;
   int          tw;
   Widget       topWid  = NULL;
   char        *tmpChar = NULL;
   int          n;
   XawTextBlock tB;
   FILE        *fp;
   int          len;
   int          as,de;
   char        *s;
   int scrollVert = False;
   int scrollHor  = False;
   char *t;

   if ( linesT )
   {
      lines = atoi(linesT);
      lines = lines < 10 ? 10 : lines;
   }
   else
      lines = 25;

   if( columnsT )
   {
      cols = atoi(columnsT);
      cols = cols <= 40 ? 40 : cols;
   }

   /***** main form ******/
   mf = createContainer(toplevel,"logForm");

   /******** title label *******/
   tlb = createTitleLabel(mf, "label", title);
   topWid = tlb;

   /********* text widget *******/
   inText = createText(mf, "text", NULL, True, False, NULL, tlb);
   topWid = inText;

   /************* ok Button ************/
   okB = createButton(mf, "okButton", logText, okCB, (XtPointer)XWCHOICE_KB,
                      topWid,
                      NULL);
   selectedWidget = okB;

   /* realize the window */
   XtRealizeWidget(toplevel);
   getSpacing(mf, &hSpace, &vSpace);

   /* read actual width of label */
   XtVaGetValues(tlb,
                 XtNwidth,       &wMax,
                 NULL);

   /* look wor geometry for button */
   XtVaGetValues(okB,
                 XtNwidth,       &w,
                 XtNborderWidth, &bwButton,
                 NULL);

   if ( w < appResourceRec.minButtonWidth )
   {
      w = appResourceRec.minButtonWidth;
   }

   switch(*appResourceRec.buttonAlignment)
   {
      case 'L':
          spc = 0;
          n = bwButton + w + bwButton + hSpace;
      break;
      default:
          n = bwButton + w + bwButton;
   }
   wMax = n > wMax ? n : wMax;

   /* calculate the width for text widget according to wanted columns */
   XtVaGetValues(inText,
                 XtNfont, &fontn,
                 XtNborderWidth, &bwText,
                 NULL);
   tw = XTextWidth(fontn, "W", 1);
   tw = tw * cols + 2 * bwText + 6 /* internal border? */;
   wMax = tw > wMax ? tw : wMax;

   /* calculate the heigth for text widget according to wanted lines */
   as = fontn->max_bounds.ascent;
   de = fontn->max_bounds.descent;
   hMax = (as+de) * lines + 2 * bwText + 14 /* scrollbar fixed*/;

   /* and set this for the text */
   setWidthAndHeight(inText, wMax-2*bwText, hMax-2*bwText);

   /* set width for the other widget */
   setWidth(tlb, wMax);

   switch(*appResourceRec.buttonAlignment)
   {
      case 'L':
        spc = hSpace;
      break;
      case 'R':
         spc = wMax + hSpace - (bwButton + w + bwButton);
      break;
      default:
         spc = hSpace + (wMax - (bwButton + w + bwButton))/2;
   }

   setWidthAndX(okB, w, spc);

   /* get height an width for main form */
   h = vSpace ;
   calcHeight(tlb, vSpace, &h);
   h  += hMax  + vSpace; /* text widget */
   calcHeight(okB, vSpace, &h);
   hMax = h;

   /* set all values */
   wMax+=hSpace*2;
   n=0;

   setShellGeoAndTitle(toplevel, wMax, hMax, title);
   XtMapWidget(toplevel);

   if ( errFile )
   {
      if ( strcmp(errFile, "-") )
         fp = fopen(errFile,"r");
      else
         fp = stdin;
      if ( fp && ! tail )
      {
         fseek(fp,0,SEEK_END);
         len = ftell(fp);
         fseek(fp,0,0);
         tmpChar = calloc(len+1,1);
         fread(tmpChar,len,1, fp);
         fclose(fp);
         size=len;
         /* count number of lines and columns
          * set if necessary the scrollbars on
          */

         s = t = tmpChar;
         while ( s && *s )
         {
            s = (char*)strchr(s, '\n');
            if ( s )
            {
               logLines++;
               logColumns = logColumns > (s-t) ? logColumns : (s-t);
               s++;
               t = s;
            }
         }
         if ( logLines > lines)  scrollVert = True;
         if ( logColumns > cols) scrollHor  = True;

         tB.firstPos = 0;
         tB.length   = len;
         tB.ptr      = tmpChar;
         tB.format   = 0;
         XtVaSetValues(inText,  XtNeditType, XawtextAppend, NULL);
         XawTextReplace(inText, 1, 1,&tB);
         free(tmpChar);
         XtVaSetValues(inText, XtNeditType, XawtextRead, NULL);
      }
      else if ( fp && tail )
      {
         n = fileno(fp);
#ifdef USE_TIMER
         XtAppAddTimeOut(appContext,10, prtText, (XtPointer)n);
#else
         XtAppAddInput(appContext,
                       n, (XtPointer) XtInputReadMask,
                       prtText, NULL );
#endif
      }
      else
      {
         fprintf(stderr,"cant open file %s, error %d\n",errFile,errno);
         exit(2);
      }

      if ( scrollVert )
      {
         XtVaSetValues(inText,
                       XtNscrollVertical, XawtextScrollAlways,
                       NULL);
 
         XtAddEventHandler(inText,
                           KeyPressMask|KeyReleaseMask|ButtonReleaseMask|ButtonPressMask|PointerMotionMask|ButtonMotionMask,
                           True,
                           (XtEventHandler) scrollMsg_CB, (XtPointer)True);
      }
      else
      {
         XtAddEventHandler(inText,
                           KeyPressMask|KeyReleaseMask|ButtonReleaseMask|ButtonPressMask|PointerMotionMask|ButtonMotionMask,
                           True,
                           (XtEventHandler) scrollMsg_CB, (XtPointer)False);
      }

      if ( scrollHor )
         XtVaSetValues(inText,
                       XtNscrollHorizontal, XawtextScrollAlways,
                       NULL);
      XtSetKeyboardFocus( toplevel, inText);
   }
   else
   {
      fprintf(stderr,"no file given\n");
      exit(2);
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void getHeight
 *
 *  FUNCTION:   read the heihht of given widget and return the
 *              new height for the window
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

static void calcHeight(Widget w, Dimension vSpace, Dimension *newHeight)
{
   Dimension h;
   Dimension bw;
   if ( w != NULL )
   {
      XtVaGetValues(w,
                    XtNheight, &h,
                    XtNborderWidth, &bw,
                    NULL);
      *newHeight += vSpace + h +bw +bw;
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void getSpacing
 *
 *  FUNCTION:   read spacing values
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

static void getSpacing(Widget w, Dimension *hor, Dimension *vert)
{
   if ( w != NULL )
   {
      XtVaGetValues(w,
                    XtNdefaultDistance, vert,
                    NULL);
      *hor = *vert;
   }
   else
   {
      *hor  = HSPACE;
      *vert = VSPACE;
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void setWidth
 *
 *  FUNCTION:   set width for the given widget
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void setWidth(Widget w, Dimension width)
{
   if ( w )
   {
      XtUnmanageChild(w);
      if ( width )
      {
         XtVaSetValues(w,
                       XtNresizable, False,
                       XtNresize, False,
                       XtNwidth,  width,
                       NULL);
         XtManageChild(w);
      }
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void setWidthAndX
 *
 *  FUNCTION:   set width for the given widget
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void setWidthAndX(Widget w, Dimension width, Dimension x)
{
   if ( w )
   {
      XtUnmanageChild(w);
      if ( width )
      {
         XtVaSetValues(w,
                       XtNresizable,     False,
                       XtNresize,        False,
                       XtNwidth,         width,
                       XtNhorizDistance, x,
                       NULL);
         XtManageChild(w);
      }
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void setWidthAndHeight
 *
 *  FUNCTION:   set with and height of widget
 *
 *  INPUT:     Widget     w
 *             Dimemsion  width
 *             Dimemsion  height
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void setWidthAndHeight(Widget w, Dimension width, Dimension height)
{
   if ( w )
   {
      XtUnmanageChild(w);
      XtVaSetValues(w,
                    XtNresizable, False,
                    XtNresize, False,
                    XtNwidth,  width,
                    XtNheight, height,
                    NULL);
      XtManageChild(w);
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       Dimension getMaxWidth
 *
 *  FUNCTION:   get with of widget and return the witdh or
 *              maxW
 *
 *  INPUT:     Widget     wid,
 *             Dimemsion  maxW
 *  OUTPUT:    dim_t     *dim
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

Dimension getMaxWidth(Widget wid, Dimension maxW, dim_t *dim)
{
   if ( wid )
   {
      XtVaGetValues(wid,
                    XtNborderWidth, &dim->bw,
                    XtNwidth,       &dim->w,
                    XtNheight,      &dim->h,
                    NULL);
      dim->tw = dim->bw + dim->bw + dim->w;
      return dim->tw > maxW ? dim->tw : maxW;
   }
   else
   {
      return maxW;
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void buildLoginWindow
 *
 *  FUNCTION:   build the login window
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void buildLoginWindow()
{
   Widget loginWindow;
   Widget loginLB  = NULL;
   Widget passwdLB = NULL;
   Widget wid = NULL;
   dim_t  dtilb;
   dim_t  dtq;
   dim_t  dlolb;
   dim_t  dpalb;
   dim_t  dloin;
   dim_t  dpain;
   Dimension w1,w2;
   Dimension dh, dv;
   Dimension  maxW = 0, h = 0;
   dim_t dca;
   dim_t dok;
   Position p;
   int n;
   Arg   args[20];

   tlb = NULL;
   loginWindow = createContainer(toplevel, "loginForm");
   
   if ( title )
   {
      wid = tlb = createTitleLabel(loginWindow, "label", title);
   }

   /**** Create question *****/
   if ( question )
   {
      n = 0;
      XtSetArg(args[n], XtNborderWidth, 0); n++;
      XtSetArg(args[n], XtNjustify,     XtJustifyLeft); n++;
      XtSetArg(args[n], XtNlabel,       question); n++;
      XtSetArg(args[n], XtNresizable,   True); n++;
      XtSetArg(args[n], XtNresize,      False); n++;
      XtSetArg(args[n], XtNtop,         XtChainTop);n++;
      XtSetArg(args[n], XtNbottom,      XtChainTop);n++;
      XtSetArg(args[n], XtNfromVert,    tlb);n++;
      XtSetArg(args[n], XtNleft,        XtChainLeft);n++;
      XtSetArg(args[n], XtNright,       XtChainLeft);n++;

      quest = XtCreateManagedWidget("question",labelWidgetClass, loginWindow, args,n);
      wid = quest;
   }

   if ( loginTxt )
   {
      loginLB = createLabel(loginWindow, "loginLB", loginTxt, wid);
      wid = loginIn = createRText(loginWindow, "text", wid, loginLB, True);
      XtAddEventHandler(wid, KeyPressMask|KeyReleaseMask|ButtonPressMask,
                        True,
                        (XtEventHandler) setFocus_CB, NULL);
      XtOverrideTranslations(loginIn, XtParseTranslationTable(textTranslationSpc));
      addWidgetTab(wid);
   }

   if ( passwdTxt )
   {
      passwdLB = createLabel(loginWindow, "passwdLB", passwdTxt, wid);
      wid = passwdIn = createRText(loginWindow, "text", wid, passwdLB, /*loginTxt?False:*/True);
      XtAddEventHandler(wid, KeyPressMask|KeyReleaseMask|ButtonPressMask, True,
                        (XtEventHandler) setFocus_CB, NULL);
      /* add callback for invisible input */
      XtAddEventHandler( passwdIn, KeyPressMask|KeyReleaseMask,
                         True,
                         (XtEventHandler) invisibleInput_CB, NULL);
      addWidgetTab(wid);
   }

   /* create OK and cancel button */
   if ( cancel )
   {
      cancelB = createButton(loginWindow, "cancel", cancel, cancelCB, NULL, wid, NULL);
      XtAddEventHandler(cancelB, KeyPressMask|KeyReleaseMask, True,
                        (XtEventHandler) setFocus_CB, NULL);
      XtOverrideTranslations(cancelB,
                             XtParseTranslationTable(buttonTranslation));
      XtAddEventHandler(cancelB, LeaveWindowMask|EnterWindowMask,  True,
                        (XtEventHandler) leave_CB, NULL);
      addWidgetTab(cancelB);
   }

   okB = createButton(loginWindow, "okButton", ok, passwdCB, (XtPointer)XWCHOICE_KB, wid, NULL);

   XtAddEventHandler(okB, KeyPressMask|KeyReleaseMask, True,
                     (XtEventHandler) setFocus_CB, NULL);
   XtOverrideTranslations(okB,
                          XtParseTranslationTable(buttonTranslation));
   XtAddEventHandler(okB, LeaveWindowMask|EnterWindowMask,  True,
                     (XtEventHandler) leave_CB, NULL);
   addWidgetTab(okB);
   XtRealizeWidget(toplevel);
   getSpacing(loginWindow, &dh, &dv);

   if ( loginTxt && passwdTxt )
   {
      XtVaGetValues(loginLB, XtNwidth, &w1, NULL);
      XtVaGetValues(passwdLB, XtNwidth, &w2, NULL);
      if ( w1 > w2 )
      {
         w2 = w1 - w2 + dh;
         XtVaSetValues(passwdIn, XtNhorizDistance, w2,  NULL);
      }
      else
      {
         w1 = w2 - w1 + dh;
         XtVaSetValues(loginIn, XtNhorizDistance, w1,  NULL);
      }
   }

   /* get size for labels and text fields */

   maxW = getMaxWidth(tlb, maxW, &dtilb);
   maxW = getMaxWidth(quest, maxW, &dtq);
   w1 = getMaxWidth(loginLB, 0, &dlolb);
   w2 = getMaxWidth(loginIn, 0, &dloin);
   w1 += w2 + dh;
   maxW = maxW > w1 ? maxW : w1;
   w1 = getMaxWidth(passwdLB, 0, &dpalb);
   w2 = getMaxWidth(passwdIn, 0, &dpain);
   w1 += w2 + dh;
   maxW = maxW > w1 ? maxW : w1;

   /* get / set size for buttons */
   getMaxWidth(okB, 0, &dok);
   getMaxWidth(cancelB, 0, &dca);
   if ( appResourceRec.minButtonWidth > 0 )
   {
      w1 = dca.w > dok.w ?  dca.w : dok.w;
      w1 = w1 > appResourceRec.minButtonWidth
            ? w1
            : appResourceRec.minButtonWidth;
      dca.w = dok.w = w1;
   }
   w1 = dca.bw + dca.w + dca.bw + dh + dok.bw + dok.w + dok.bw;
   maxW = maxW > w1 ? maxW : w1;
   
   switch(*appResourceRec.buttonAlignment)
   {
      case 'L':
         setWidthAndX(cancelB, dca.w, dh);
         setWidthAndX(okB, dca.w,  dh + dca.bw + dca.w + dca.bw + dh);
      break;
      case 'R':
         w1 = maxW + dh - dok.bw - dok.w -dok.bw;
         setWidthAndX(okB, dok.w, w1);
         w1 -= dh + dca.bw + dca.w + dca.bw;
         setWidthAndX(cancelB, dca.w, w1);
      break;
      case 'c': /* center */
         w1 = maxW/2 + dh; /* center of window */
         w2 = w1 - dh/2 - dca.w - dca.bw - dca.bw;
         setWidthAndX(cancelB, dca.w, w2);
         w2 = w1 + dh/2;
         setWidthAndX(okB, dok.w, w2);
      break;
      default:  /* Center, use the whole width */
         setWidthAndX(cancelB, dca.w, dh);
         w1 = maxW + dh - dok.bw - dok.w -dok.bw;
         setWidthAndX(okB, dok.w, w1);
   }

   setWidth(tlb,maxW);
   setWidth(quest,maxW);
   if ( loginIn )
   {
      XtVaGetValues(loginIn, XtNx, &p, NULL);
   }
   else
   {
      XtVaGetValues(passwdIn, XtNx, &p, NULL);
   }
   w1 = maxW - p;
   setWidth(loginIn, w1); 
   setWidth(passwdIn, w1); 

   h = dv;
   if ( tlb ) h += dtilb.h + dv;
   if ( quest ) h += dtq.h + dv;
   if ( loginIn ) h += dloin.h + dloin.bw*2 + dv;
   if ( passwdIn ) h += dpain.h + dpain.bw*2 + dv;
   h += dok.h + dok.bw*2 + dv;

   setShellGeoAndTitle(toplevel, maxW+dh+dh, h, title);
   XtMapWidget(toplevel);
   wid = loginIn ? loginIn : passwdIn;
   XtSetKeyboardFocus( toplevel, wid);
   if ( wid != passwdIn )
   {
      XawTextDisplayCaret(passwdIn, False);
   }
   XtVaSetValues(wid, XtNborderColor, darkblue, NULL);
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void buildCheckWindow
 *
 *  FUNCTION:   build the checkbox window
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void buildCheckWindow()
{
   Widget mf;
   Widget wid;
   Dimension vSpace, hSpace,w1, w2, h;
   int   n;
   Arg   args[20];
   check_t *c;
   Dimension maxW = 0;
   Dimension w = 0;
   dim_t dok;
   dim_t dca;
   Position p;

   /***** main form ******/
   mf = createContainer(toplevel, "choiceForm");

   /******** title label *******/
   tlb = createTitleLabel(mf, "label", title);
   wid = tlb;

   XtRealizeWidget(toplevel);
   getSpacing(mf, &hSpace, &vSpace);

   XtVaGetValues(tlb, XtNwidth, &maxW, NULL);

   /**** Create question *****/
   if ( question )
   {
      n = 0;
      XtSetArg(args[n], XtNborderWidth, 0); n++;
      XtSetArg(args[n], XtNjustify,     XtJustifyLeft); n++;
      XtSetArg(args[n], XtNlabel,       question); n++;
      XtSetArg(args[n], XtNresizable,   True); n++;
      XtSetArg(args[n], XtNresize,      False); n++;
      XtSetArg(args[n], XtNtop,         XtChainTop);n++;
      XtSetArg(args[n], XtNbottom,      XtChainTop);n++;
      XtSetArg(args[n], XtNfromVert,    tlb);n++;
      XtSetArg(args[n], XtNleft,        XtChainLeft);n++;
      XtSetArg(args[n], XtNright,       XtChainLeft);n++;

      quest = XtCreateManagedWidget("question",labelWidgetClass, mf, args,n);
      wid = quest;
      XtVaGetValues(quest, XtNwidth, &w, NULL);
      maxW = maxW > w ? maxW : w;
   }

   c = check;
   while(c)
   {
      c->box = wid = createCheck(mf, wid, c);
      maxW = maxW > c->width ? maxW : c->width;
      if ( defCheck && strcmp(defCheck, c->text) == 0 )
      {
         XtVaSetValues(c->box, XtNstate,1, NULL);
      }
      addWidgetTab(c->box);
      XtAddEventHandler(c->box, KeyPressMask|KeyReleaseMask, True,
                        (XtEventHandler) setFocus_CB, NULL);
      XtAddEventHandler(c->box, EnterWindowMask|LeaveWindowMask,  True,
                        (XtEventHandler) leave_CB, NULL);
      XtOverrideTranslations(c->box,
                             XtParseTranslationTable(toggleTranslation));
      c = c->next;
   }

   if ( defCheck == NULL && radio )
   {
      XtVaSetValues(check->box,
                    XtNstate, 1,
                    NULL); 
   }

   /* create OK and cancel button */
   if ( cancel )
   {
      cancelB = createButton(mf, "cancel", cancel, cancelCB, NULL, wid, NULL);
      XtAddEventHandler(cancelB, KeyPressMask|KeyReleaseMask, True,
                        (XtEventHandler) setFocus_CB, NULL);
      XtOverrideTranslations(cancelB,
                             XtParseTranslationTable(buttonTranslation));
      XtAddEventHandler(cancelB, LeaveWindowMask|EnterWindowMask,  True,
                        (XtEventHandler) leave_CB, NULL);
      addWidgetTab(cancelB);
   }

   okB = createButton(mf, "okButton", ok, checkOkCB, (XtPointer)XWCHOICE_KB, wid, NULL);

   XtAddEventHandler(okB, KeyPressMask|KeyReleaseMask, True,
                     (XtEventHandler) setFocus_CB, NULL);
   XtOverrideTranslations(okB,
                          XtParseTranslationTable(buttonTranslation));
   XtAddEventHandler(okB, LeaveWindowMask|EnterWindowMask,  True,
                     (XtEventHandler) leave_CB, NULL);
   addWidgetTab(okB);

   /* get / set size for buttons */
   getMaxWidth(okB, 0, &dok);
   getMaxWidth(cancelB, 0, &dca);
   if ( appResourceRec.minButtonWidth > 0 )
   {
      w1 = dca.w > dok.w ?  dca.w : dok.w;
      w1 = w1 > appResourceRec.minButtonWidth
            ? w1
            : appResourceRec.minButtonWidth;
      dca.w = dok.w = w1;
   }
   w1 = dca.bw + dca.w + dca.bw + hSpace + dok.bw + dok.w + dok.bw;
   maxW = maxW > w1 ? maxW : w1;
   
   switch(*appResourceRec.buttonAlignment)
   {
      case 'L':
         setWidthAndX(cancelB, dca.w, hSpace);
         setWidthAndX(okB, dca.w,  hSpace + dca.bw + dca.w + dca.bw + hSpace);
      break;
      case 'R':
         w1 = maxW + hSpace - dok.bw - dok.w -dok.bw;
         setWidthAndX(okB, dok.w, w1);
         w1 -= hSpace + dca.bw + dca.w + dca.bw;
         setWidthAndX(cancelB, dca.w, w1);
      break;
      case 'c': /* center */
         w1 = maxW/2 + hSpace; /* center of window */
         w2 = w1 - hSpace/2 - dca.w - dca.bw - dca.bw;
         setWidthAndX(cancelB, dca.w, w2);
         w2 = w1 + hSpace/2;
         setWidthAndX(okB, dok.w, w2);
      break;
      default:  /* Center, use the whole width */
         setWidthAndX(cancelB, dca.w, hSpace);
         w1 = maxW + hSpace - dok.bw - dok.w -dok.bw;
         setWidthAndX(okB, dok.w, w1);
   }

   setWidth(tlb,maxW);
   XtVaGetValues(okB,
                 XtNy, &p,
                 XtNheight, &h,
                 NULL);

   p = p + dok.bw + h +  hSpace;
                 
   setShellGeoAndTitle(toplevel, maxW+hSpace*2, p, title);
   XtMapWidget(toplevel);
   XtSetKeyboardFocus( toplevel, check->box);
   XtCallAcceptFocus(check->box,CurrentTime);
   XtVaSetValues(check->box, XtNborderColor, darkblue, NULL);
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void buildWindow
 *
 *  FUNCTION:   build the main window
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

void buildWindow()
{
   Widget       wid;
   Arg          args[20];
   XrmValue     fromVal, toVal;
   int          n = 0;
   int          i;
   char        *s, *t;
   Dimension    w, h, rowSpc;
   Dimension    x1, x2;
   Dimension    wMax;
   XFontStruct *fontb=NULL, *fontn=NULL;
   int          tw, th;
   int          cols = 1;
   int          defaultChoice = 0;
   Widget       topWid  = NULL;
   int          moreLines = 1;
   dim_t        dim[OKBUTTON+1];
   static       char linesText[100];
   Dimension    vSpace = 4;
   Dimension    hSpace = 4;

   memset(dim, 0, sizeof(dim));

   if ( linesT )
   {
      lines = atoi(linesT);
      moreLines = 0;
   }
   if ( lines == 0 )
      lines = 15;

   if( columnsT )
   {
      cols = atoi(columnsT);
   }
   cols = cols == 0 ? 1 : cols;


   /******************************************************/
   /* get fonts                                          */
   /******************************************************/

   /* standard font for list and buttons */
   wid = XtCreateManagedWidget("tmp",labelWidgetClass,toplevel,NULL,0);
   XtVaGetValues(wid,
                 XtNfont, &fontn,
                 NULL);
   XtDestroyWidget(wid);

   /* bold bigger font for titel an question */

   if ( fb )
   {
      fromVal.addr = fb;
      toVal.addr = NULL;
      fromVal.size = sizeof(String);
      XtConvert(toplevel, XtRString, (XrmValuePtr) &fromVal, XtRFontStruct,
                   (XrmValuePtr) &toVal);
      if ( toVal.addr != NULL )
      {
         fontb = (XFontStruct*) *((XFontStruct **)toVal.addr);
      }
   }
   else
   {
      /* from resource XwChoice*label.font */
      fontb = fontn;
   }

   /******************************************************/
   /* calculate text length in pixel                     */
   /******************************************************/

   /* this may be a multi line string ! */
   s = t = question;
   do
   {
      t = strchr(s,'\n');
      if ( t )
      {
         tw = XTextWidth(fontb,s, t-s);
         dim[QUESTION].w = tw > dim[QUESTION].w ? tw : dim[LABEL].w ;
         s = t+1;
      }
      else
      {
         tw = XTextWidth(fontb,s, strlen(s));
         dim[QUESTION].w = tw > dim[QUESTION].w ? tw : dim[LABEL].w ;
      }
   } while ( t );

   /* TBD */
   dim[QUESTION].w += 4; /* otherwise rightmost half char would be cut out -
                            but really looks there's too offset at left ...*/

   tw = 0;
   for ( i = 0; i < choice.count; i++)
   {
      tw = XTextWidth(fontn,choice.text[i], strlen(choice.text[i]));
      if ( cols == 1 )
      {
         tw += 2;
         dim[LIST].w = tw > dim[LIST].w ? tw : dim[LIST].w ;
      }
      else
      {
         tw = (tw+8)*cols;
         dim[LIST].w = tw > dim[LIST].w ? tw : dim[LIST].w ;
      }
      if(defChoice && strcmp(defChoice,choice.text[i]) == 0 )
      {
         /* default choice */
         if ( defChoice && strcmp(choice.text[i],defChoice) == 0)
         {
            defaultChoice = i;
         }
      }
   }

   if ( defaultChoice == choice.count )
   {
      defaultChoice = 0;
   }

   /***** main form ******/
   mf = createContainer(toplevel, "choiceForm");

   /******** title label *******/
   tlb = createTitleLabel(mf, "label", title);
   topWid = tlb;

   XtRealizeWidget(toplevel);
   getSpacing(mf, &hSpace, &vSpace);

   /**** Create question *****/
   n = 0;
   XtSetArg(args[n], XtNfont,        fontb);n++;
   XtSetArg(args[n], XtNborderWidth, 0); n++;
   XtSetArg(args[n], XtNjustify,     XtJustifyLeft); n++;
   XtSetArg(args[n], XtNlabel,       question); n++;
   XtSetArg(args[n], XtNresizable,   True); n++;
   XtSetArg(args[n], XtNresize,      False); n++;
   XtSetArg(args[n], XtNtop,         XtChainTop);n++;
   XtSetArg(args[n], XtNbottom,      XtChainTop);n++;
   XtSetArg(args[n], XtNfromVert,    tlb);n++;
   XtSetArg(args[n], XtNleft,        XtChainLeft);n++;
   XtSetArg(args[n], XtNright,       XtChainLeft);n++;

   quest = XtCreateManagedWidget("question",labelWidgetClass, mf, args,n);
   topWid = quest;

   /**** Create List *****/
   if ( choice.count )
   {
      th = fontn->max_bounds.ascent + fontn->max_bounds.descent + 2;
      if ( lines > ((choice.count+cols-1)/cols) )
      {
         lines = (choice.count+cols-1)/cols;
         linesT = linesText;
         sprintf(linesText, "%d",lines);
      }

      n = 0;
      /* preset the width and height of the view port */
      XtSetArg(args[n], XtNtop,       XtChainTop);n++;
      XtSetArg(args[n], XtNbottom,    XtChainTop);n++;
      XtSetArg(args[n], XtNfromVert,  topWid);n++;
      XtSetArg(args[n], XtNleft,      XtChainLeft);n++;
      XtSetArg(args[n], XtNright,     XtChainLeft);n++;
      XtSetArg(args[n], XtNheight,    20); n++;
      XtSetArg(args[n], XtNwidth,     20);   n++;
      XtSetArg(args[n], XtNresize,    False); n++;
      XtSetArg(args[n], XtNresizable, True); n++;
      vp = XtCreateManagedWidget("vp",viewportWidgetClass,mf,args,n);
      /* get scrollbar and set borderWidth to 1 ! */

      viewPortHeight = th * lines + 4;
      topWid = vp;

      n = 0;
      XtSetArg(args[n], XtNdefaultColumns, cols); n++;
      XtSetArg(args[n], XtNforceColumns,   True); n++;
      XtSetArg(args[n], XtNverticalList,   True); n++;
      XtSetArg(args[n], XtNresize,         True); n++;
      XtSetArg(args[n], XtNresizable,      True); n++;
      XtSetArg(args[n], XtNrowSpacing,     2); n++;
      XtSetArg(args[n], XtNinternalHeight, 2); n++;
      XtSetArg(args[n], XtNinternalWidth,  2); n++;
      XtSetArg(args[n], XtNborderWidth,    0); n++;

      /* set the choices */
      if ( choice.count )
      {
         XtSetArg(args[n], XtNlist, choice.text); n++;
      }

      /* create button as given choice in a list */
      list = XtCreateManagedWidget("list",listWidgetClass,
                                   vp, args,n);
      XtAddEventHandler(list,
                        KeyPressMask|KeyReleaseMask|ButtonReleaseMask|ButtonPressMask,
                        True,
                        (XtEventHandler) ListSel_CB, NULL);
      XtAddEventHandler(list, StructureNotifyMask,
                        True,
                       (XtEventHandler) setFocus_CB, NULL);
      addWidgetTab(list);
   }

   if ( input )
   {
      inText = createText(mf, "text", NULL, False, True, NULL, topWid);
      addWidgetTab(inText);
      XtAddEventHandler(inText, KeyPressMask|KeyReleaseMask|ButtonPressMask|ButtonReleaseMask,
                        True,
                        (XtEventHandler) setFocus_CB, NULL);

      XtOverrideTranslations(inText,
                             Input == 0
                                ?
                                  XtParseTranslationTable(textTranslation)
                                :
                                  XtParseTranslationTable(textTranslationSpc));
      topWid = inText;
   }

   /******* install the help, cancel and ok buttons */
   if ( cancel )
   {
      cancelB = createButton(mf, "cancel", cancel, cancelCB, NULL, topWid, NULL);
      addWidgetTab(cancelB);
      XtAddEventHandler(cancelB, KeyPressMask|KeyReleaseMask, True,
                        (XtEventHandler) setFocus_CB, NULL);
      XtOverrideTranslations(cancelB,
                             XtParseTranslationTable(buttonTranslation));
      XtAddEventHandler(cancelB, LeaveWindowMask|EnterWindowMask,  True,
                        (XtEventHandler) leave_CB, NULL);
   }

   if ( help )
   {
      helpB = createButton(mf, "help", help, helpCB, NULL, topWid, NULL);
      addWidgetTab(helpB);
      XtAddEventHandler(helpB, KeyPressMask|KeyReleaseMask, True,
                        (XtEventHandler) setFocus_CB, NULL);
      XtOverrideTranslations(helpB,
                             XtParseTranslationTable(buttonTranslation));
      XtAddEventHandler(helpB, LeaveWindowMask|EnterWindowMask,  True,
                        (XtEventHandler) leave_CB, NULL);
      XtVaGetValues(list, XtNrowSpacing, &rowSpc, NULL);
   }

   if ( ok )
   {
      okB = createButton(mf, "okButton", ok, okCB, (XtPointer)XWCHOICE_KB, topWid, NULL);

      addWidgetTab(okB);
      XtAddEventHandler(okB, KeyPressMask|KeyReleaseMask, True,
                        (XtEventHandler) setFocus_CB, NULL);
      XtOverrideTranslations(okB,
                             XtParseTranslationTable(buttonTranslation));
      XtAddEventHandler(okB, LeaveWindowMask|EnterWindowMask,  True,
                        (XtEventHandler) leave_CB, NULL);

   }

   /* calculate and set the geometry */
   wMax= 0;
   getMaxWidth(cancelB, wMax, &dim[CANCEL]);
   getMaxWidth(okB,     wMax, &dim[OKBUTTON]);
   getMaxWidth(helpB,   wMax, &dim[HELP]);

   /* calculate height */
   if ( choice.count )
   {
      int scrollbarVertical   = 0;
      int scrollbarHorizontal = 0;
      /* calculate now the size for our viewport and list */
      /* if the hight is to big for our screen we have to */
      /* st this accordingly with consideration of the    */
      /* scrollbar                                        */
      /* make shure that the height will not be to big */

      th = fontn->max_bounds.ascent + fontn->max_bounds.descent + 2;
      if ( (lines * th)+ h >
           screenH - appResourceRec.rigthMargin - appResourceRec.leftMargin )
      {
         if ( moreLines )
         {
            lines = (screenH - appResourceRec.topMargin - appResourceRec.bottomMargin -100 )/th;
            linesT = linesText;
            sprintf(linesText, "%d",lines);
         }
         scrollbarVertical = 1;
      }

      if ( lines > ((choice.count+cols-1)/cols) )
      {
         if ( moreLines )
         {
            lines = (choice.count+cols-1)/cols;
            linesT = linesText;
            sprintf(linesText, "%d",lines);
         }
         scrollbarVertical = 1;
      }

      if ( lines < ((choice.count+cols-1)/cols) )
      {
         if ( moreLines )
         {
            lines = ((choice.count+cols-1)/cols);
            linesT = linesText;
            sprintf(linesText, "%d",lines);
            scrollbarVertical = 0;
         }
         else
         {
            scrollbarVertical = 1;
         }
      }

      if ( scrollbarVertical )
      {
         dim[LIST].w += SCRBW; /* for the scrollbar */
      }

      if ( dim[LIST].w > screenW -
                        appResourceRec.leftMargin -
                        appResourceRec.rigthMargin )
      {
         scrollbarHorizontal = 1;
         wMax = screenW - appResourceRec.leftMargin -
                          appResourceRec.rigthMargin - 8;
      }
      else
      {
         wMax = wMax > dim[LIST].w ? wMax : dim[LIST].w;
      }

      XtVaGetValues(vp, XtNborderWidth, &dim[VIEWPORT].bw, NULL);
      dim[VIEWPORT].w = dim[LIST].w;
/* TBD */
      dim[VIEWPORT].h = th * lines + 7 + scrollbarHorizontal * SCRBW;

      if ( scrollbarVertical )
      {
         XtVaSetValues(vp,
                       XtNallowVert, True,
                       XtNuseRight,  True,
                       NULL);
      }
      if ( scrollbarHorizontal )
      {
         XtVaSetValues(vp,
                       XtNallowHoriz, True,
                       XtNuseBottom,  True,
                       NULL);
      }
   }

   /* height for list is known, calculate the window height */
   h = vSpace;

   calcHeight(tlb, vSpace, &h);
   calcHeight(quest, vSpace, &h);
   h += dim[VIEWPORT].h + dim[VIEWPORT].bw*2 + vSpace;
   calcHeight(inText, vSpace, &h);
   calcHeight(okB, vSpace, &h);

   /* calc max and min width / height for the toplevel shell */
   wMax = dim[LABEL].bw * 2 + dim[LABEL].w;
   if ( choice.count )
   {
      if ( lines < (choice.count+cols-1)/cols)
         dim[LIST].w += SCRBW;
      w = dim[LIST].bw * 2 + dim[LIST].w + 10;
      wMax = wMax > w ? wMax : w;
   }

   if ( wMax + 8 > screenW - appResourceRec.rigthMargin - appResourceRec.leftMargin )
   {
      wMax = screenW - appResourceRec.rigthMargin - appResourceRec.leftMargin - 20;
   }
   wMax = getMaxWidth(tlb,    wMax, &dim[LABEL]);
   wMax = getMaxWidth(quest,  wMax, &dim[QUESTION]);
   wMax = getMaxWidth(inText, wMax, &dim[TEXT]);

   /*arrangement of buttons, OK will always be here
    * all buttons -> cancel spacer1 help spacer2 ok
    * cancel + ok -> cancel spacer1 ok
    */

   /* button width, all with equal width (minButtonWidth > 0) */
   if ( appResourceRec.minButtonWidth > 0 )
   {
      w = dim[OKBUTTON].w > dim[CANCEL].w ? dim[OKBUTTON].w : dim[CANCEL].w;
      w = w > appResourceRec.minButtonWidth ? w : appResourceRec.minButtonWidth;
      w = dim[HELP].w > w ? dim[HELP].w : w;
      dim[OKBUTTON].w = dim[CANCEL].w = dim[HELP].w = w;
   }

   n = 1;
   /* with of button array */
   if ( helpB == NULL )
   {
      w = dim[CANCEL].bw + dim[CANCEL].w + dim[CANCEL].bw +
          hSpace +
          dim[OKBUTTON].bw + dim[OKBUTTON].w + dim[OKBUTTON].bw;
   }
   else /* all buttons */
   {
      w = dim[CANCEL].bw + dim[CANCEL].w + dim[CANCEL].bw +
          hSpace +
          dim[HELP].bw + dim[HELP].w + dim[HELP].bw +
          hSpace +
          dim[OKBUTTON].bw + dim[OKBUTTON].w + dim[OKBUTTON].bw;
   }
   wMax = w > wMax ? w : wMax;

   /* set width for all widgets */
   setWidth(tlb, wMax);
   setWidth(vp, wMax-dim[LIST].bw*2);

   /* visible area within view port */
   viewPortWidth = wMax-dim[LIST].bw*2;

   /* if there is a vertical scrollbar */
   if ((choice.count+cols-1)/cols > lines)
   {
      viewPortWidth  -= SCRBW;
   }

   setWidth(list,    wMax-dim[LIST].bw*2);
   setWidth(quest,   wMax);
   setWidthAndHeight(vp, wMax-dim[VIEWPORT].bw*2, dim[VIEWPORT].h );
   setWidth(inText,  wMax-dim[TEXT].bw*2);
   switch(*appResourceRec.buttonAlignment)
   {
      case 'L':
         w = hSpace;
         setWidthAndX(cancelB, dim[CANCEL].w, w);
         w += dim[CANCEL].bw + dim[CANCEL].w + dim[CANCEL].bw + hSpace;
         if ( helpB )
         {
            setWidthAndX(helpB, dim[HELP].w, w);
            w += dim[HELP].bw  + dim[HELP].w + dim[HELP].bw + hSpace;
         }
         setWidthAndX(okB, dim[OKBUTTON].w, w);
      break;
      case 'R':
         w = wMax + hSpace - dim[OKBUTTON].bw - dim[OKBUTTON].w - dim[OKBUTTON].bw;
         setWidthAndX(okB, dim[OKBUTTON].w, w);
         if ( helpB )
         {
            w -= hSpace + dim[HELP].bw + dim[HELP].w + dim[HELP].bw;
            setWidthAndX(helpB, dim[HELP].w, w);
         }
         w -= hSpace +  dim[CANCEL].w + dim[CANCEL].bw*2;
         setWidthAndX(cancelB, dim[CANCEL].w, w);
      break;
      case 'c': /* center */
         x1 = x2 = hSpace + wMax/2;
         if ( helpB )
         {
            x1 -=  dim[HELP].w / 2 + dim[HELP].bw;
            x2 +=  dim[HELP].w / 2 + dim[HELP].bw;
            setWidthAndX(helpB, dim[HELP].w, x1);
            x1 -= hSpace/2;
            x2 += hSpace/2;
         }
         w = x1 - hSpace/2 - dim[CANCEL].bw - dim[CANCEL].w - dim[CANCEL].bw;
         setWidthAndX(cancelB, dim[CANCEL].w, w);
         x2 += hSpace/2;
         setWidthAndX(okB, dim[OKBUTTON].w, x2);
         break;
      default:  /* Center, use the whole width */
         setWidthAndX(cancelB, dim[CANCEL].w, hSpace);
         x1 = hSpace + dim[CANCEL].w + dim[CANCEL].bw*2;

         x2 = wMax + hSpace - dim[OKBUTTON].w - dim[OKBUTTON].bw * 2;
         setWidthAndX(okB, dim[OKBUTTON].w, x2);

         w = x1 + (x2 - x1 - dim[HELP].w)/2 - dim[HELP].bw;
         setWidthAndX(helpB, dim[HELP].w, w);
   }   

   /* set values for the shell */
   setShellGeoAndTitle(toplevel, wMax+hSpace*2, h, title);
   XtMapWidget(toplevel);

   /* select first list item or default choice */
   if ( list && !noReturn )
   {
      XawListHighlight(list, defaultChoice);
      setListPosition(list, defaultChoice, XWCHOICE_KB, KeyPress);
   }

   if ( list )
   {
      XtSetKeyboardFocus( toplevel, list);
      XtCallAcceptFocus(list,CurrentTime);
      XtVaSetValues(XtParent(list), XtNborderColor, darkblue, NULL);
   }
   else if ( inText )
   {
      XtSetKeyboardFocus( toplevel, inText);
      XtCallAcceptFocus(inText,CurrentTime);
      XtVaSetValues(inText, XtNborderColor, darkblue, NULL);
   }
}

/*
 *********************************************************************
 *
 *  NAME:       usage
 *
 *********************************************************************
 */

void usage (int e, char *opt)
{
   puts(xwChoice_Vxc);
   puts(WITHXAW);
   if ( e == 2 ) return;
   if ( e == 1 )
   {
     fputs("Error: not enough arguments, or wrong format:",stderr);
     if ( opt )
     {
        fputs(opt,stderr);
     }
     fputs("\n",stderr);
   }
   fputs("Usage: ", stdout);
   fputs(program, stdout);
   fputs(" options\n\n"
     "  where options are (check manual/helpfile for details):\n"
     "    -display display\n"
     "    -columns #\n"
     "    -lines #\n"
     "    -geometry XxY[+x+y]\n"
     "    -title titleString\n"
     "    -help helpButtonText helpOK helpFile\n"
     "    -cancel cancelButtonText cancelAnswer\n"
     "    -ok okButtonText\n"
     "    -question questionString\n"
     "    -choice choiceN answerN\n"
     "    -defChoice defChoice\n"
     "    -check checkN answerN\n"
     "    -defCheck defCheck\n"
     "    -error okButtonText errorFile\n"
     "    -message okButtonText messageFile\n"
     "    -wait OkButtonString waitMessageText\n"
     "    -input\n"
     "    -Input\n"
     "    -deny denyChars\n"
     "    -both\n"
     "    -h|--help\n"
     "    -noReturn\n"
     "    -log okButtonText  logFile\n"
     "    -tail okButtonText  logFile\n"
     "    -login loginText\n"
     "    -passwd passwdText\n"
     "    -v|--version\n\n",
     stdout);
}


/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void mainLoop
 *
 *  FUNCTION:   process X events and dispatch them
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

typedef struct name_s
{
   struct name_s *next;
   char *name;
} name_t;

void mainLoop(void)
{
   Widget w;
   Widget p;
   Display *display = XtDisplay(toplevel);
   XEvent event;

   memset(&event, 0, sizeof(event));

   for (;;)
   {
      XtAppNextEvent(appContext,&event);
      if ( event.xclient.type == ClientMessage )
      {
         break;
      }

      w = XtWindowToWidget(display , event.xany.window);
      p = XtParent(w);

      if ( (event.xany.type == ButtonPress || event.xany.type == ButtonRelease )
            &&
            (event.xbutton.button == 4 || event.xbutton.button == 5 )
          )
      {
         /* redirect srollbar events to the text or list widget */
         if (strcmp("vScrollbar", XtName(w)) == 0 )
         {
            event.xany.window = XtWindow(p);
         }
         else if (strcmp("vertical", XtName(w)) == 0)
         {
            event.xany.window = XtWindow(list);
         }
      }

      /* handle Paste for the password widget */
      if ( w == passwdIn )
      {
         Time   time;
         switch(event.type)
         {
             case ButtonPress:
                if ( event.xbutton.button == Button2 )
                {
                   /* button 2 pressed */
                   time = XtLastTimestampProcessed( XtDisplay( toplevel ));
                   XtGetSelectionValue(passwdIn, XA_PRIMARY, XA_STRING, 
                                       (XtSelectionCallbackProc)do_selection_callback, NULL, time );
                   /* disallow Xaw-buildin paste */
                   event.xbutton.button = Button1; 
                }
                break;
             case SelectionNotify:
                 /* now do_selection_callback will be called */
                 break;;
             case SelectionClear:
             case SelectionRequest:
                continue;
         }
      }

      XtDispatchEvent(&event);
   }
}

/*
 *********************************************************************
 *  +++fhb
 *
 *  NAME:       void main
 *
 *  FUNCTION:   start here
 *
 *  INPUT:
 *
 *  REMARKS:
 *
 *  +++fhe
 *********************************************************************
*/

int main(int argc, char *argv[])
{
   char    *args[20];
   int      count;
   char     *s, *t;
   char     *location = NULL;
   char     *center  = NULL;
   XrmDatabase database;

   args[0] = *argv;
   args[1] = "-display";
   args[3] = NULL;
   args[4] = NULL;
   args[5] = NULL;
   args[6] = NULL;
   args[7] = NULL;
   args[8] = NULL;
   args[9] = NULL;
   args[10] = NULL;
   args[11] = NULL;
   args[12] = NULL;
   count   = 3;

   /* preset title to prg name */
   s = strrchr(*argv,'/');
   program = (s!=NULL) ? s+1 : *argv;
   title   = program;

   s = getenv("DISPLAY");
   if ( s != NULL )
      display = s;

   fn = getenv("XwChoiceFn");
   fb = getenv("XwChoiceFb");

   argc--;
   argv++;
   while ( argc > 0 )
   {
      IF_OPT("-display",        display)
      else IF_OPT("-columns",   columnsT)
      else IF_OPT("-lines",     linesT)
      else IF_OPT("-geometry",  geometry)
      else IF_OPT("-location",  location)
      else IF_OPT("-center",    center)
      else IF_OPT("-fn",        fn)
      else IF_OPT("-fb",        fb)
      else IF_OPT("-title",     title)
      else IF_OPT("-login",     loginTxt)
      else IF_OPT("-passwd",    passwdTxt)
      else IF_OPTH("-help",     help, helpOK, helpFile)
      else IF_OPTB("-cancel",   cancel, cancelRet)
      else IF_OPT("-ok",        ok)
      else IF_OPT("-question",  question)
      else IF_OPTCc("-check",   check)
      else IF_OPTC("-choice",   choice)
      else IF_OPT("-defChoice", defChoice)
      else IF_OPT("-defCheck",  defCheck)
      else IF_OPTB("-error",    error, errFile)
      else IF_OPTB("-message",  message, errFile)
      else IF_OPTB("-wait",     Wait, errFile)
      else IF_OPT1("-input",    input)
      else IF_OPT1("-Input",    Input)
      else IF_OPT("-deny",      denyChars)
      else IF_OPT1("-both",     both)
      else IF_OPT1("-radio",    radio)
      else IF_OPT1("-noReturn", noReturn)
      else IF_OPTB("-log",      logText, errFile )
      else IF_OPTB("-tail",     tail,errFile )
      else if (strcmp(*argv,"-h") == 0 || strcmp(*argv,"--help") == 0)
      {
         usage(0, NULL);
         exit(0);
      }
      else if (strcmp(*argv,"-v") == 0 || strcmp(*argv,"--version") == 0)
      {
         usage(2, NULL);
         exit(0);
      }
      else
      {
         snprintf(errorText, sizeof(errorText)," Unknown option %s\n", *argv);
         usage(1, *argv);
         exit(2);
      }
   }

   if ( Input == 1 || both == 1 )
      input = 1;

   if ( tail != NULL )
   {
      logText = tail;
   }

   if ( (choice.text != NULL && input == 0 ) ||
        (input != 0 && choice.text == NULL &&
         Wait == NULL && logText == NULL ) )
   {
      both = 0;
   }

   if ( question == NULL && message == NULL && error == NULL && Wait == NULL
        && choice.text == NULL && logText == NULL && input == 0 && both == 0
        && loginTxt == NULL && passwdTxt == NULL && check == NULL )
   {
      usage(1, NULL);
      exit(2);
   }

   if ( question == NULL && input != 0 )
   {
      usage(1, "expect -question <question text>");
      exit(2);
   }
   
   /***************************************************/
   /* substitute \n with <NL> (question string)       */
   /***************************************************/
   if ( question )
   {
      t = s = question;
      while(*s)
      {
         if ( *s == '\\' && s[1] == 'n' )
         {
            *t++ = '\n';
            s += 2;
         }
         else
            *t++ = *s++;
      }
      *t = '\0';
   }
   else
   {
      if ( choice.text != NULL || check != NULL )
      {
        if ( both  ) question = "choose and enter text:";
        else if ( input ) question = "choose or enter text:";
        else question = "choose:";
      }
      else
        if ( input ) question = "enter text:";
   }

   args[2] = display;
   if ( geometry )
   {
      s = geometry;
      if ( *s )
      {
         args[count++] = "-geometry";
         args[count++] = geometry;
      }
   }

   if ( fn )
   {
      args[count++] = "-font";
      args[count++] = fn;
   }
   
   toplevel = XtVaAppInitialize(&appContext,  /* XtAppContext */
                                class,        /* String       */
                                NULL,         /* XrmOptionDescRec* */
                                0,            /* Cardinal num_options */
                                &count,       /* int* argc_in_out; */
                                args,         /* String* argv_in_out */
                                fallback,     /* String* fallback_resources */
                                NULL
                               );

   if ( toplevel )
   {
      Window win;
      Atom XwAtom = None;
      Bool oie;

      /* screen width and height */
      screenH = DisplayHeight(XtDisplay(toplevel),0);
      screenW = DisplayWidth(XtDisplay(toplevel),0);
      XtVaSetValues(toplevel, XtNmappedWhenManaged, False, NULL);
      getColors();

      /* get our special resources */
      XtGetApplicationResources(toplevel, (XtPointer)&appResourceRec,
                                resources, XtNumber(resources),
                                NULL, 0);

      /* overload a few values, not modifiable by extarnal resources! */
      
      database = XtDatabase(XtDisplay(toplevel));
      XrmPutLineResource(&database, "xwChoice.logForm.text.vScrollbar.width: 13");
      XrmPutLineResource(&database, "xwChoice.logForm.text.vScrollbar.borderWidth: 1");
      XrmPutLineResource(&database, "xwChoice.logForm.text.hScrollbar.height:13");
      XrmPutLineResource(&database, "xwChoice.logForm.text.hScrollbar.borderWidth:1");
      
      XrmPutLineResource(&database, "xwChoice.error.text.vScrollbar.width: 13");
      XrmPutLineResource(&database, "xwChoice.error.text.vScrollbar.borderWidth: 1");
      XrmPutLineResource(&database, "xwChoice.error.text.hScrollbar.height:13");
      XrmPutLineResource(&database, "xwChoice.error.text.hScrollbar.borderWidth:1");
      XrmPutLineResource(&database, "XwChoice*error*vScrollbar*shadowWidth:  2");
      XrmPutLineResource(&database, "XwChoice*error*hScrollbar*shadowWidth:  2");
      XrmPutLineResource(&database, "XwChoice*error.okButton.shadowWidth:  2");
      XrmPutLineResource(&database, "XwChoice*error*shadowWidth:  0");
      
      XrmPutLineResource(&database, "*.helpForm.text.vScrollbar.width: 13");
      XrmPutLineResource(&database, "*.helpForm.text.vScrollbar.borderWidth: 1");
      XrmPutLineResource(&database, "*.helpForm.text.hScrollbar.height: 13");
      XrmPutLineResource(&database, "*.helpForm.text.hScrollbar.borderWidth: 1");
      
      XrmPutLineResource(&database, "xwChoice.message.text.vScrollbar.width: 13");
      XrmPutLineResource(&database, "xwChoice.message.text.vScrollbar.borderWidth: 1");
      XrmPutLineResource(&database, "xwChoice.message.text.hScrollbar.height: 13");
      XrmPutLineResource(&database, "xwChoice.message.text.hScrollbar.borderWidth: 1");
      XrmPutLineResource(&database, "XwChoice*message*vScrollbar*shadowWidth:  2");
      XrmPutLineResource(&database, "XwChoice*message*hScrollbar*shadowWidth:  2");
      XrmPutLineResource(&database, "XwChoice*message.okButton.shadowWidth:  2");
      XrmPutLineResource(&database, "XwChoice*message*shadowWidth:  0");
      
      XrmPutLineResource(&database, "xwChoice.choiceForm.vp.vertical.width: 13");
      XrmPutLineResource(&database, "xwChoice.choiceForm.vp.vertical.borderWidth: 1");
      XrmPutLineResource(&database, "xwChoice.choiceForm.vp.horizontal.height: 13");
      XrmPutLineResource(&database, "xwChoice.choiceForm.vp.horizontal.borderWidth: 1");
      XrmPutLineResource(&database, "xwChoice.choiceForm.vp.horizontal.shadowWidth: 2");
      XrmPutLineResource(&database, "xwChoice.choiceForm.vp.vertical.shadowWidth: 2");

      XrmPutLineResource(&database, "xwChoice*helpForm*vScrollbar.width: 13");
      XrmPutLineResource(&database, "xwChoice*helpForm*vScrollbar.borderWidth: 1");
      XrmPutLineResource(&database, "xwChoice*helpForm*hScrollbar.height:13");
      XrmPutLineResource(&database, "xwChoice*helpForm*hScrollbar.borderWidth:1");
      XrmPutLineResource(&database, "XwChoice*helpForm*vScrollbar*shadowWidth:  2");
      XrmPutLineResource(&database, "XwChoice*helpForm*hScrollbar*shadowWidth:  2");
      XrmPutLineResource(&database, "XwChoice*helpForm.okButton.shadowWidth: 2");
      XrmPutLineResource(&database, "XwChoice*helpForm*shadowWidth:  0");

      if ( center && (*center == '0' || *center == 'f' ||  *center == 'F') )
      {
         appResourceRec.center = 0;
      }

      if ( location )
      {
         appResourceRec.location = location;
      }

      if ( check )
      {
         buildCheckWindow();
      }
      else if ( loginTxt || passwdTxt )
      {
         buildLoginWindow();
      }
      else if ( error == NULL && message == NULL &&
           Wait == NULL && logText == NULL )
      {
         buildWindow();
      }
      else
      {
         if ( logText )
         {
            buildLog();
         }
         else
         {
            showError();
         }
      }

      oie = Wait != NULL ? False : True;
      XwAtom = XInternAtom(XtDisplay(toplevel),
                           class,
                           oie);

      if ( XwAtom != None )
      {
         if ( (win = XGetSelectionOwner(XtDisplay(toplevel), XwAtom)) == None )
         {
            if ( Wait )
            {
               XSetSelectionOwner(XtDisplay(toplevel),XwAtom,
                                  XtWindow(toplevel), CurrentTime);
            }
         }

         if ( win != None && Wait == NULL )
         {
            /* send a client message to the Wait window */
            XEvent event;

            memset(&event,0,sizeof(XEvent));
            event.xclient.type = ClientMessage;
            event.xclient.window = win;
            event.xclient.display = XtDisplay(toplevel);
            event.xclient.message_type = None;
            event.xclient.format = 8;
            XSendEvent(XtDisplay(toplevel), win, True,NoEventMask,&event);
         }
         else if ( Wait != NULL && win != None && win != XtWindow(toplevel))
         {
            exit(0);
         }
      }
      mainLoop();
   }
   else
   {
      fputs(program, stderr);
      fputs(": error: could not init Xt-toplevel.\n", stderr);
      exit(1);
   }
   exit(0);
   return 0;
}
